# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini, Albert Chen, Kaixu Yang, Sayan Patra
"""Functions to generate derived time features useful
in forecasting, such as growth, seasonality, holidays.
"""

import inspect
import math
import warnings
from datetime import datetime

import fbprophet.hdays as fbholidays
import holidays
import numpy as np
import pandas as pd
from scipy.special import expit

from greykite.common import constants as cst


def convert_date_to_continuous_time(dt):
    """Converts date to continuous time. Each year is one unit.

    Parameters
    ----------
    dt : datetime object
        the date to convert

    Returns
    -------
    conti_date : `float`
        the date represented in years
    """
    year_length = datetime(dt.year, 12, 31).timetuple().tm_yday
    tt = dt.timetuple()

    return (dt.year +
            (tt.tm_yday - 1
             + dt.hour / 24
             + dt.minute / (24 * 60)
             + dt.second / (24 * 3600)) / float(year_length))


def get_default_origin_for_time_vars(df, time_col):
    """Sets default value for origin_for_time_vars

    Parameters
    ----------
    df : `pandas.DataFrame`
        Training data. A data frame which includes the timestamp and value columns
    time_col : `str`
        The column name in `df` representing time for the time series data.

    Returns
    -------
    dt_continuous_time : `float`
        The time origin used to create continuous variables for time
    """
    date = pd.to_datetime(df[time_col][0])
    return convert_date_to_continuous_time(date)


def build_time_features_df(dt, conti_year_origin):
    """This function gets a datetime-like vector and creates new columns containing temporal
    features useful for time series analysis and forecasting e.g. year, week of year, etc.

    Parameters
    ----------
    dt : array-like (1-dimensional)
        A vector of datetime-like values
    conti_year_origin : float
        The origin used for creating continuous time.

    Returns
    -------
    time_features_df : `pandas.DataFrame`
        Dataframe with the following time features.

            * "datetime": `datetime.datetime` object, a combination of date and a time
            * "date": `datetime.date` object, date with the format (year, month, day)
            * "year": integer, year of the date e.g. 2018
            * "year_length": integer, number of days in the year e.g. 365 or 366
            * "quarter": integer, quarter of the date, 1, 2, 3, 4
            * "quarter_start": `pandas.DatetimeIndex`, date of beginning of the current quarter
            * "quarter_length": integer, number of days in the quarter, 90/91 for Q1, 91 for Q2, 92 for Q3 and Q4
            * "month": integer, month of the year, January=1, February=2, ..., December=12
            * "month_length": integer, number of days in the month, 28/ 29/ 30/ 31
            * "woy": integer, ISO 8601 week of the year where a week starts from Monday, 1, 2, ..., 53
            * "doy": integer, ordinal day of the year, 1, 2, ..., year_length
            * "doq": integer, ordinal day of the quarter, 1, 2, ..., quarter_length
            * "dom": integer, ordinal day of the month, 1, 2, ..., month_length
            * "dow": integer, day of the week, Monday=1, Tuesday=2, ..., Sunday=7
            * "str_dow": string, day of the week as a string e.g. "1-Mon", "2-Tue", ..., "7-Sun"
            * "str_doy": string, day of the year e.g. "2020-03-20" for March 20, 2020
            * "hour": integer, discrete hours of the datetime, 0, 1, ..., 23
            * "minute": integer, minutes of the datetime, 0, 1, ..., 59
            * "second": integer, seconds of the datetime, 0, 1, ..., 3599
            * "year_month": string, (year, month) e.g. "2020-03" for March 2020
            * "year_woy": string, (year, week of year) e.g. "2020_42" for 42nd week of 2020
            * "month_dom": string, (month, day of month) e.g. "02/20" for February 20th
            * "year_woy_dow": string, (year, week of year, day of week) e.g. "2020_03_6" for Saturday of 3rd week in 2020
            * "woy_dow": string, (week of year, day of week) e.g. "03_6" for Saturday of 3rd week
            * "dow_hr": string, (day of week, hour) e.g. "4_09" for 9am on Thursday
            * "dow_hr_min": string, (day of week, hour, minute) e.g. "4_09_10" for 9:10am on Thursday
            * "tod": float, time of day, continuous, 0.0 to 24.0
            * "tow": float, time of week, continuous, 0.0 to 7.0
            * "tom": float, standardized time of month, continuous, 0.0 to 1.0
            * "toq": float, time of quarter, continuous, 0.0 to 1.0
            * "toy": float, standardized time of year, continuous, 0.0 to 1.0
            * "conti_year": float, year in continuous time, eg 2018.5 means middle of the year 2018
            * "is_weekend": boolean, weekend indicator, True for weekend, else False
            * "dow_grouped": string, Monday-Thursday=1234-MTuWTh, Friday=5-Fri, Saturday=6-Sat, Sunday=7-Sun
            * "ct1": float, linear growth based on conti_year_origin, -infinity to infinity
            * "ct2": float, signed quadratic growth, -infinity to infinity
            * "ct3": float, signed cubic growth, -infinity to infinity
            * "ct_sqrt": float, signed square root growth, -infinity to infinity
            * "ct_root3": float, signed cubic root growth, -infinity to infinity

    """
    dt = pd.DatetimeIndex(dt)
    if len(dt) == 0:
        raise ValueError("Length of dt cannot be zero.")

    # basic time features
    date = dt.date
    year = dt.year
    year_length = (365.0 + dt.is_leap_year)
    quarter = dt.quarter
    month = dt.month
    month_length = dt.days_in_month

    # finds first day of quarter
    quarter_start = pd.DatetimeIndex(
        dt.year.map(str) + "-" + (3 * quarter - 2).map(int).map(str) + "-01")
    next_quarter_start = dt + pd.tseries.offsets.QuarterBegin(startingMonth=1)
    quarter_length = (next_quarter_start - quarter_start).days
    # finds offset from first day of quarter (rounds down to nearest day)
    doq = ((dt - quarter_start) / pd.to_timedelta("1D") + 1).astype(int)

    # week of year, "woy", follows ISO 8601:
    #   - Week 01 is the week with the year's first Thursday in it.
    #   - A week begins with Monday and ends with Sunday.
    # So the week number of the week that overlaps both years, is 1, 52, or 53,
    # depending on whether it has more days in the previous year or new year.
    #   - e.g. Jan 1st, 2018 is Monday. woy of first 8 days = [1, 1, 1, 1, 1, 1, 1, 2]
    #   - e.g. Jan 1st, 2019 is Tuesday. woy of first 8 days = [1, 1, 1, 1, 1, 1, 2, 2]
    #   - e.g. Jan 1st, 2020 is Wednesday. woy of first 8 days = [1, 1, 1, 1, 1, 2, 2, 2]
    #   - e.g. Jan 1st, 2015 is Thursday. woy of first 8 days = [1, 1, 1, 1, 2, 2, 2, 2]
    #   - e.g. Jan 1st, 2021 is Friday. woy of first 8 days = [53, 53, 53, 1, 1, 1, 1, 1]
    #   - e.g. Jan 1st, 2022 is Saturday. woy of first 8 days = [52, 52, 1, 1, 1, 1, 1, 1]
    #   - e.g. Jan 1st, 2023 is Sunday. woy of first 8 days = [52, 1, 1, 1, 1, 1, 1, 1]
    woy = dt.strftime("%V").astype(int)
    doy = dt.dayofyear
    dom = dt.day
    dow = dt.strftime("%u").astype(int)
    str_dow = dt.strftime("%u-%a")  # e.g. 1-Mon, 2-Tue, ..., 7-Sun
    hour = dt.hour
    minute = dt.minute
    second = dt.second

    # grouped time feature
    str_doy = dt.strftime("%Y-%m-%d")       # e.g. 2020-03-20 for March 20, 2020
    year_month = dt.strftime("%Y-%m")       # e.g. 2020-03 for March 2020
    month_dom = dt.strftime("%m/%d")        # e.g. 02/20 for February 20th
    year_woy = dt.strftime("%Y_%V")         # e.g. 2020_42 for 42nd week of 2020
    year_woy_dow = dt.strftime("%Y_%V_%u")  # e.g. 2020_03_6 for Saturday of 3rd week in 2020
    woy_dow = dt.strftime("%W_%u")          # e.g. 03_6 for Saturday of 3rd week
    dow_hr = dt.strftime("%u_%H")           # e.g. 4_09 for 9am on Thursday
    dow_hr_min = dt.strftime("%u_%H_%M")    # e.g. 4_09_10 for 9:10am on Thursday

    # derived time features
    tod = hour + (minute / 60.0) + (second / 3600.0)
    tow = dow - 1 + (tod / 24.0)
    tom = (dom - 1 + (tod / 24.0)) / month_length
    toq = (doq - 1 + (tod / 24.0)) / quarter_length
    # time of year, continuous, 0.0 to 1.0. e.g. Jan 1, 12 am = 0/365, Jan 2, 12 am = 1/365, ...
    # To handle leap years, Feb 28 = 58/365 - 59/365, Feb 29 = 59/365, Mar 1 = 59/365 - 60/365
    # offset term is nonzero only in leap years
    # doy_offset reduces doy by 1 from from Mar 1st (doy > 60)
    doy_offset = (year_length == 366) * 1.0 * (doy > 60)
    # tod_offset sets tod to 0 on Feb 29th (doy == 60)
    tod_offset = 1 - (year_length == 366) * 1.0 * (doy == 60)
    toy = (doy - 1 - doy_offset + (tod / 24.0) * tod_offset) / 365.0

    # year of date in continuous time, eg 2018.5 means middle of year 2018
    # this is useful for modeling features that do not care about leap year e.g. environmental variables
    conti_year = year + (doy - 1 + (tod / 24.0)) / year_length
    is_weekend = pd.Series(dow).apply(lambda x: x in [6, 7]).values  # weekend indicator
    # categorical var with levels (Mon-Thu, Fri, Sat, Sun), could help when training data are sparse.
    dow_grouped = pd.Series(str_dow).apply(lambda x: "1234-MTuWTh" if (x in ["1-Mon", "2-Tue", "3-Wed", "4-Thu"]) else x).values

    # growth terms
    ct1 = conti_year - conti_year_origin
    ct2 = signed_pow(ct1, 2)
    ct3 = signed_pow(ct1, 3)
    ct_sqrt = signed_pow(ct1, 1/2)
    ct_root3 = signed_pow(ct1, 1/3)

    # All keys must be added to constants.
    features_dict = {
        "datetime": dt,
        "date": date,
        "year": year,
        "year_length": year_length,
        "quarter": quarter,
        "quarter_start": quarter_start,
        "quarter_length": quarter_length,
        "month": month,
        "month_length": month_length,
        "woy": woy,
        "doy": doy,
        "doq": doq,
        "dom": dom,
        "dow": dow,
        "str_dow": str_dow,
        "str_doy": str_doy,
        "hour": hour,
        "minute": minute,
        "second": second,
        "year_month": year_month,
        "year_woy": year_woy,
        "month_dom": month_dom,
        "year_woy_dow": year_woy_dow,
        "woy_dow": woy_dow,
        "dow_hr": dow_hr,
        "dow_hr_min": dow_hr_min,
        "tod": tod,
        "tow": tow,
        "tom": tom,
        "toq": toq,
        "toy": toy,
        "conti_year": conti_year,
        "is_weekend": is_weekend,
        "dow_grouped": dow_grouped,
        "ct1": ct1,
        "ct2": ct2,
        "ct3": ct3,
        "ct_sqrt": ct_sqrt,
        "ct_root3": ct_root3,
    }
    df = pd.DataFrame(features_dict)
    return df


def add_time_features_df(df, time_col, conti_year_origin):
    """Adds a time feature data frame to a data frame
    :param df: the input data frame
    :param time_col: the name of the time column of interest
    :param conti_year_origin: the origin of time for the continuous time variable
    :return: the same data frame (df) augmented with new columns
    """
    df = df.reset_index(drop=True)
    time_df = build_time_features_df(
        dt=df[time_col],
        conti_year_origin=conti_year_origin)
    time_df = time_df.reset_index(drop=True)
    return pd.concat([df, time_df], axis=1)


def get_holidays(countries, year_start, year_end):
    """This function extracts a holiday data frame for the period of interest
    [year_start to year_end] for the given countries.
    This is done using the holidays libraries in pypi:fbprophet and pypi:holidays
    Implementation resembles that of `~fbprophet.make_holidays.make_holidays_df`

    Parameters
    ----------
    countries : `list` [`str`]
        countries for which we need holidays
    year_start : `int`
        first year of interest, inclusive
    year_end : `int`
        last year of interest, inclusive

    Returns
    -------
    holiday_df_dict : `dict` [`str`, `pandas.DataFrame`]
        - key: country name
        - value: data frame with holidays for that country
          Each data frame has two columns: EVENT_DF_DATE_COL, EVENT_DF_LABEL_COL
    """
    country_holiday_dict = {}
    year_list = list(range(year_start, year_end + 1))

    for country in countries:
        try:
            # Fetch the holidays from fbprophet holiday set
            # Suppress the following warning for India:
            #   "We only support Diwali and Holi holidays from 2010 to 2025"
            if country in ["India", "IN"]:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    country_holidays = getattr(fbholidays, country)(years=year_list)
            else:
                country_holidays = getattr(fbholidays, country)(years=year_list)
        except AttributeError:
            # Fetch the holidays from pypi:holidays set
            try:
                country_holidays = getattr(holidays, country)(years=year_list)
            except AttributeError:
                raise AttributeError(f"Holidays in {country} are not currently supported!")

        country_df = pd.DataFrame({
            cst.EVENT_DF_DATE_COL: list(country_holidays.keys()),
            cst.EVENT_DF_LABEL_COL: list(country_holidays.values())})
        country_df[cst.EVENT_DF_DATE_COL] = pd.to_datetime(country_df[cst.EVENT_DF_DATE_COL])
        country_holiday_dict[country] = country_df

    return country_holiday_dict


def get_available_holiday_lookup_countries(countries=None):
    """Returns list of available countries for modeling holidays

    :param countries: List[str]
        only look for available countries in this set
    :return: List[str]
        list of available countries for modeling holidays
    """
    fb_countries = [
        name for name, obj in inspect.getmembers(fbholidays)
        if inspect.isclass(obj) and obj.__module__ == fbholidays.__name__]
    holidays_countries = [
        name for name, obj in inspect.getmembers(holidays)
        if inspect.isclass(obj) and obj.__module__ == holidays.__name__]
    all_countries = set(fb_countries + holidays_countries)

    if countries is not None:
        countries = set(countries)
        found_countries = all_countries.intersection(countries)
    else:
        found_countries = all_countries
    found_countries.discard("HolidayBase")  # edge case, remove if found
    return sorted(list(found_countries))


def get_available_holidays_in_countries(
        countries,
        year_start,
        year_end):
    """Returns a dictionary mapping each country to its holidays
        between the years specified.

    :param countries: List[str]
        countries for which we need holidays
    :param year_start: int
        first year of interest
    :param year_end: int
        last year of interest
    :return: Dict[str, List[str]]
        key: country name
        value: list of holidays in that country between [year_start, year_end]
    """
    country_holiday_dict = get_holidays(countries, year_start, year_end)
    country_holiday_list = {country: list(sorted(set(df[cst.EVENT_DF_LABEL_COL].values)))
                            for country, df in country_holiday_dict.items()}
    return country_holiday_list


def get_available_holidays_across_countries(
        countries,
        year_start,
        year_end):
    """Returns a list of holidays that occur any of the countries
    between the years specified.

    :param countries: List[str]
        countries for which we need holidays
    :param year_start: int
        first year of interest
    :param year_end: int
        last year of interest
    :return: List[str]
        names of holidays in any of the countries between [year_start, year_end]
    """
    country_holiday_list = get_available_holidays_in_countries(
        countries=countries,
        year_start=year_start,
        year_end=year_end)
    holiday_across_countries = {
        holiday for country, holiday_list in country_holiday_list.items()
        for holiday in holiday_list}
    return list(sorted(holiday_across_countries))


def add_daily_events(
        df,
        event_df_dict,
        date_col=cst.EVENT_DF_DATE_COL,
        regular_day_label=cst.EVENT_DEFAULT):
    """For each key of event_df_dict, it adds a new column to a data frame (df)
        with a date column (date_col).
        Each new column will represent the events given for that key.

    Notes
    -----
    As a side effect, the columns in ``event_df_dict`` are renamed.

    Parameters
    ----------
    df : `pandas.DataFrame`
        The data frame which has a date column.
    event_df_dict : `dict` [`str`, `pandas.DataFrame`]
        A dictionary of data frames, each representing events data
        for the corresponding key.
        Values are DataFrames with two columns:

            - The first column contains the date. Must be at the same
              frequency as ``df[date_col]`` for proper join. Must be in a
              format recognized by `pandas.to_datetime`.
            - The second column contains the event label for each date

    date_col : `str`
        Column name in ``df`` that contains the dates for joining against
        the events in ``event_df_dict``.
    regular_day_label : `str`
        The label used for regular days which are not "events".

    Returns
    -------
    df_daily_events : `pandas.DataFrame`
        An augmented data frame version of df with new label columns --
        one for each key of ``event_df_dict``.
    """
    df[date_col] = pd.to_datetime(df[date_col])
    for label, event_df in event_df_dict.items():
        event_df = event_df.copy()
        new_col = f"{cst.EVENT_PREFIX}_{label}"
        event_df.columns = [date_col, new_col]
        event_df[date_col] = pd.to_datetime(event_df[date_col])
        df = df.merge(event_df, on=date_col, how="left")
        df[new_col] = df[new_col].fillna(regular_day_label)

    return df


def add_event_window(
        df,
        time_col,
        label_col,
        time_delta="1D",
        pre_num=1,
        post_num=1,
        events_name=""):
    """For a data frame of events with a time_col and label_col
        it adds shifted events
        prior and after the given events
        For example if the event data frame includes the row
            '2019-12-25, Christmas'
        the function will produce dataframes with the events:
            '2019-12-24, Christmas' and '2019-12-26, Christmas'
        if pre_num and post_num are 1 or more.
    :param df: pd.DataFrame
        the events data frame with two columns 'time_col' and 'label_col'
    :param time_col: str
        The column with the timestamp of the events.
        This can be daily but does not have to
    :param label_col: str
        the column with labels for the events
    :param time_delta: str
        the amount of the shift for each unit specified by a string
        e.g. "1D" stands for one day delta
    :param pre_num: int
        the number of events to be added prior to the given event for each event in df
    :param post_num: int
        the number of events to be added after to the given event for each event in df
    :param events_name: str
        for each shift, we generate a new data frame
        and those data frames will be stored in a dictionary with appropriate keys.
        Each key starts with "events_name"
        and follow up with:
            "_minus_1", "_minus_2", "_plus_1", "_plus_2", ...
        depending on pre_num and post_num
    :return: dict[key: pd.Dataframe]
        A dictionary of dataframes for each needed shift.
        For example if pre_num=2 and post_num=3.
        2 + 3 = 5 data frames will be stored in the return dictionary.
        """

    df_dict = {}
    pd_time_delta = pd.to_timedelta(time_delta)
    for num in range(pre_num):
        df0 = pd.DataFrame()
        df0[time_col] = df[time_col] - (num + 1) * pd_time_delta
        df0[label_col] = df[label_col]
        df_dict[events_name + "_minus_" + f"{(num + 1):.0f}"] = df0

    for num in range(post_num):
        df0 = pd.DataFrame()
        df0[time_col] = df[time_col] + (num + 1) * pd_time_delta
        df0[label_col] = df[label_col]
        df_dict[events_name + "_plus_" + f"{(num + 1):.0f}"] = df0

    return df_dict


def get_evenly_spaced_changepoints_values(
        df,
        continuous_time_col="ct1",
        n_changepoints=2):
    """Partitions interval into n_changepoints + 1 segments,
        placing a changepoint at left endpoint of each segment.
        The left most segment doesn't get a changepoint.
        Changepoints should be determined from training data.

    :param df: pd.DataFrame
        training dataset. contains continuous_time_col
    :param continuous_time_col: str
        name of continuous time column (e.g. conti_year, ct1)
    :param n_changepoints: int
        number of changepoints requested
    :return: np.array
        values of df[continuous_time_col] at the changepoints
    """
    if not n_changepoints > 0:
        raise ValueError("n_changepoints must be > 0")

    n = df.shape[0]
    n_steps = n_changepoints + 1
    step_size = n / n_steps
    indices = np.floor(np.arange(start=1, stop=n_steps) * step_size)
    return df[continuous_time_col][indices].values


def get_evenly_spaced_changepoints_dates(
        df,
        time_col,
        n_changepoints):
    """Partitions interval into n_changepoints + 1 segments,
        placing a changepoint at left endpoint of each segment.
        The left most segment doesn't get a changepoint.
        Changepoints should be determined from training data.

    :param df: pd.DataFrame
        training dataset. contains continuous_time_col
    :param time_col: str
        name of time column
    :param n_changepoints: int
        number of changepoints requested
    :return: pd.Series
        values of df[time_col] at the changepoints
    """
    if not n_changepoints >= 0:
        raise ValueError("n_changepoints must be >= 0")
    changepoint_indices = np.floor(np.arange(start=1, stop=n_changepoints + 1) * (df.shape[0] / (n_changepoints + 1)))
    changepoint_indices = df.index[np.concatenate([[0], changepoint_indices.astype(int)])]
    return df.loc[changepoint_indices, time_col]


def get_custom_changepoints_values(
        df,
        changepoint_dates,
        time_col=cst.TIME_COL,
        continuous_time_col="ct1"):
    """Returns the values of continuous_time_col at the
        requested changepoint_dates.

    :param df: pd.DataFrame
        training dataset. contains continuous_time_col and time_col
    :param changepoint_dates: Iterable[Union[int, float, str, datetime]]
        Changepoint dates, interpreted by pd.to_datetime.
        Changepoints are set at the closest time on or after these dates
        in the dataset
    :param time_col: str
        The column name in `df` representing time for the time series data
        The time column can be anything that can be parsed by pandas DatetimeIndex
    :param continuous_time_col: str
        name of continuous time column (e.g. conti_year, ct1)
    :return: np.array
        values of df[continuous_time_col] at the changepoints
    """
    ts = pd.to_datetime(df[time_col])
    changepoint_dates = pd.to_datetime(changepoint_dates)
    # maps each changepoint to first date >= changepoint in the dataframe
    # if there is no such date, the changepoint is dropped (it would not be useful anyway)
    changepoint_ts = [ts[ts >= date].min() for date in changepoint_dates if any(ts >= date)]
    indices = ts.isin(changepoint_ts)
    changepoints = df[indices][continuous_time_col].values
    if changepoints.shape[0] == 0:
        changepoints = None
    return changepoints


def get_changepoint_string(changepoint_dates):
    """Gets proper formatted strings for changepoint dates.

    The default format is "_%Y_%m_%d_%H". When necessary, it appends "_%M" or "_%M_%S".

    Parameters
    ----------
    changepoint_dates : `list`
        List of changepoint dates, parsable by `pandas.to_datetime`.

    Returns
    -------
    date_strings : `list[`str`]`
        List of string formatted changepoint dates.
    """
    changepoint_dates = list(pd.to_datetime(changepoint_dates))
    time_format = "_%Y_%m_%d_%H"
    if any([stamp.second != 0 for stamp in changepoint_dates]):
        time_format += "_%M_%S"
    elif any([stamp.minute != 0 for stamp in changepoint_dates]):
        time_format += "_%M"
    date_strings = [date.strftime(time_format) for date in changepoint_dates]
    return date_strings


def get_changepoint_features(
        df,
        changepoint_values,
        continuous_time_col="ct1",
        growth_func=None,
        changepoint_dates=None):
    """Returns features for growth terms with continuous time origins at
        the changepoint_values (locations) specified

    Generates a time series feature for each changepoint:
        Let t = continuous_time value, c = changepoint value
        Then the changepoint feature value at time point t is
            `growth_func(t - c) * I(t >= c)`, where I is the indicator function
        This represents growth as a function of time, where the time origin is
        the changepoint

    In the typical case where growth_func(0) = 0 (has origin at 0),
        the total effect of the changepoints is continuous in time.
        If `growth_func` is the identity function, and `continuous_time`
        represents the year in continuous time, these terms form the basis for a
        continuous, piecewise linear curve to the growth trend.
        Fitting these terms with linear model, the coefficents represent slope
        change at each changepoint

    Intended usage
    ----------

    To make predictions (on test set)
        Allow growth term as a function of time to change at these points.

    Parameters
    ----------
    :param df: pd.Dataframe
        The dataset to make predictions. Contains column continuous_time_col.
    :param changepoint_values: array-like
        List of changepoint values (on same scale as df[continuous_time_col]).
        Should be determined from training data
    :param continuous_time_col: Optional[str]
        Name of continuous time column in df
        growth_func is applied to this column to generate growth term
        If None, uses "ct1", linear growth
    :param growth_func: Optional[callable]
        Growth function for defining changepoints (scalar -> scalar).
        If None, uses identity function to use continuous_time_col directly
        as growth term
    :param changepoint_dates: Optional[list]
        List of change point dates, parsable by `pandas.to_datetime`.
    :return: pd.DataFrame, shape (df.shape[0], len(changepoints))
        Changepoint features, 0-indexed
    """
    if continuous_time_col is None:
        continuous_time_col = "ct1"
    if growth_func is None:
        def growth_func(x):
            return x

    if changepoint_dates is not None:
        time_postfixes = get_changepoint_string(changepoint_dates)
    else:
        time_postfixes = [""] * len(changepoint_values)

    changepoint_df = pd.DataFrame()
    for i, changepoint in enumerate(changepoint_values):
        time_feature = np.array(df[continuous_time_col]) - changepoint  # shifted time column (t - c_i)
        growth_term = np.array([growth_func(max(x, 0)) for x in time_feature])  # growth as a function of time
        time_feature_ind = time_feature >= 0  # Indicator(t >= c_i), lets changepoint take effect starting at c_i
        new_col = growth_term * time_feature_ind
        new_changepoint = pd.Series(new_col, name=f"{cst.CHANGEPOINT_COL_PREFIX}{i}{time_postfixes[i]}")
        changepoint_df = pd.concat([changepoint_df, new_changepoint], axis=1)
    return changepoint_df


def get_changepoint_values_from_config(
        changepoints_dict,
        time_features_df,
        time_col=cst.TIME_COL):
    """Applies the changepoint method specified in `changepoints_dict` to return the changepoint values

    :param changepoints_dict: Optional[Dict[str, any]]
        Specifies the changepoint configuration.
        "method": str
            The method to locate changepoints. Valid options:
                "uniform". Places n_changepoints evenly spaced changepoints to allow growth to change.
                "custom". Places changepoints at the specified dates.
            Additional keys to provide parameters for each particular method are described below.
        "continuous_time_col": Optional[str]
            Column to apply `growth_func` to, to generate changepoint features
            Typically, this should match the growth term in the model
        "growth_func": Optional[func]
            Growth function (scalar -> scalar). Changepoint features are created
            by applying `growth_func` to "continuous_time_col" with offsets.
            If None, uses identity function to use `continuous_time_col` directly
            as growth term
        If changepoints_dict["method"] == "uniform", this other key is required:
            "n_changepoints": int
                number of changepoints to evenly space across training period
        If changepoints_dict["method"] == "custom", this other key is required:
            "dates": Iterable[Union[int, float, str, datetime]]
                Changepoint dates. Must be parsable by pd.to_datetime.
                Changepoints are set at the closest time on or after these dates
                in the dataset.
    :param time_features_df: pd.Dataframe
        training dataset. contains column "continuous_time_col"
    :param time_col: str
        The column name in `time_features_df` representing time for the time series data
        The time column can be anything that can be parsed by pandas DatetimeIndex
        Used only in the "custom" method.
    :return: np.array
        values of df[continuous_time_col] at the changepoints
    """
    changepoint_values = None
    if changepoints_dict is not None:
        valid_changepoint_methods = ["uniform", "custom"]
        changepoint_method = changepoints_dict.get("method")
        continuous_time_col = changepoints_dict.get("continuous_time_col")

        if changepoint_method is None:
            raise Exception("changepoint method must be specified")

        if changepoint_method not in valid_changepoint_methods:
            raise NotImplementedError(
                f"changepoint method {changepoint_method} not recognized. "
                f"Must be one of {valid_changepoint_methods}")

        if changepoint_method == "uniform":
            if changepoints_dict["n_changepoints"] > 0:
                params = {"continuous_time_col": continuous_time_col} if continuous_time_col is not None else {}
                changepoint_values = get_evenly_spaced_changepoints_values(
                    df=time_features_df,
                    n_changepoints=changepoints_dict["n_changepoints"],
                    **params)
        elif changepoint_method == "custom":
            params = {}
            if time_col is not None:
                params["time_col"] = time_col
            if continuous_time_col is not None:
                params["continuous_time_col"] = continuous_time_col
            changepoint_values = get_custom_changepoints_values(
                df=time_features_df,
                changepoint_dates=changepoints_dict["dates"],
                **params)

    return changepoint_values


def get_changepoint_features_and_values_from_config(
        df,
        time_col,
        changepoints_dict=None,
        origin_for_time_vars=None):
    """Extracts changepoints from changepoint configuration and input data

    :param df: pd.DataFrame
        Training data. A data frame which includes the timestamp and value columns
    :param time_col: str
        The column name in `df` representing time for the time series data
        The time column can be anything that can be parsed by pandas DatetimeIndex
    :param changepoints_dict: Optional[Dict[str, any]]
        Specifies the changepoint configuration.
        "method": str
            The method to locate changepoints. Valid options:
                "uniform". Places n_changepoints evenly spaced changepoints to allow growth to change.
                "custom". Places changepoints at the specified dates.
            Additional keys to provide parameters for each particular method are described below.
        "continuous_time_col": Optional[str]
            Column to apply `growth_func` to, to generate changepoint features
            Typically, this should match the growth term in the model
        "growth_func": Optional[func]
            Growth function (scalar -> scalar). Changepoint features are created
            by applying `growth_func` to "continuous_time_col" with offsets.
            If None, uses identity function to use `continuous_time_col` directly
            as growth term
        If changepoints_dict["method"] == "uniform", this other key is required:
            "n_changepoints": int
                number of changepoints to evenly space across training period
        If changepoints_dict["method"] == "custom", this other key is required:
            "dates": Iterable[Union[int, float, str, datetime]]
                Changepoint dates. Must be parsable by pd.to_datetime.
                Changepoints are set at the closest time on or after these dates
                in the dataset.
    :param origin_for_time_vars: Optional[float]
        The time origin used to create continuous variables for time
    :return: Dict[str, any]
        Dictionary with the requested changepoints and associated information
        changepoint_df: pd.DataFrame, shape (df.shape[0], len(changepoints))
            Changepoint features for modeling the training data
        changepoint_values: array-like
            List of changepoint values (on same scale as df[continuous_time_col])
            Can be used to generate changepoints for prediction.
        continuous_time_col: Optional[str]
            Name of continuous time column in df
            growth_func is applied to this column to generate growth term.
            If None, uses "ct1", linear growth
            Can be used to generate changepoints for prediction.
        growth_func: Optional[callable]
            Growth function for defining changepoints (scalar -> scalar).
            If None, uses identity function to use continuous_time_col directly
            as growth term.
            Can be used to generate changepoints for prediction.
        changepoint_cols: List[str]
            Names of the changepoint columns for modeling
    """
    # extracts changepoint values
    if changepoints_dict is None:
        changepoint_values = None
        continuous_time_col = None
        growth_func = None
    else:
        if origin_for_time_vars is None:
            origin_for_time_vars = get_default_origin_for_time_vars(df, time_col)
        time_features_df = build_time_features_df(
            df[time_col],
            conti_year_origin=origin_for_time_vars)

        changepoint_values = get_changepoint_values_from_config(
            changepoints_dict=changepoints_dict,
            time_features_df=time_features_df,
            time_col="datetime")  # datetime column generated by `build_time_features_df`
        continuous_time_col = changepoints_dict.get("continuous_time_col")
        growth_func = changepoints_dict.get("growth_func")

    # extracts changepoint column names
    if changepoint_values is None:
        changepoint_df = None
        changepoint_cols = []
    else:
        if changepoints_dict is None:
            changepoint_dates = None
        elif changepoints_dict["method"] == "custom":
            changepoint_dates = list(pd.to_datetime(changepoints_dict["dates"]))
        elif changepoints_dict["method"] == "uniform":
            changepoint_dates = get_evenly_spaced_changepoints_dates(
                df=df,
                time_col=time_col,
                n_changepoints=changepoints_dict["n_changepoints"]
            ).tolist()[1:]  # the changepoint features does not include the growth term
        else:
            changepoint_dates = None
        changepoint_df = get_changepoint_features(
            df=time_features_df,
            changepoint_values=changepoint_values,
            continuous_time_col=continuous_time_col,
            growth_func=growth_func,
            changepoint_dates=changepoint_dates)
        changepoint_cols = list(changepoint_df.columns)

    return {
        "changepoint_df": changepoint_df,
        "changepoint_values": changepoint_values,
        "continuous_time_col": continuous_time_col,
        "growth_func": growth_func,
        "changepoint_cols": changepoint_cols
    }


def get_changepoint_dates_from_changepoints_dict(
        changepoints_dict,
        df=None,
        time_col=None):
    """Gets the changepoint dates from ``changepoints_dict``

    Parameters
    ----------
    changepoints_dict : `dict` or `None`
        The ``changepoints_dict`` which is compatible with
        `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`
    df : `pandas.DataFrame` or `None`, default `None`
        The data df to put changepoints on.
    time_col : `str` or `None`, default `None`
        The column name of time column in ``df``.

    Returns
    -------
    changepoint_dates : `list`
        List of changepoint dates.
    """
    if (changepoints_dict is None
            or "method" not in changepoints_dict.keys()
            or changepoints_dict["method"] not in ["auto", "uniform", "custom"]):
        return None
    method = changepoints_dict["method"]
    if method == "custom":
        # changepoints_dict["dates"] is `Iterable`, converts to list
        changepoint_dates = list(changepoints_dict["dates"])
    elif method == "uniform":
        if df is None or time_col is None:
            raise ValueError("When the method of ``changepoints_dict`` is 'uniform', ``df`` and "
                             "``time_col`` must be provided.")
        changepoint_dates = get_evenly_spaced_changepoints_dates(
            df=df,
            time_col=time_col,
            n_changepoints=changepoints_dict["n_changepoints"]
        )
        # the output is `pandas.Series`, converts to list
        changepoint_dates = changepoint_dates.tolist()[1:]
    else:
        raise ValueError("The method of ``changepoints_dict`` can not be 'auto'. "
                         "Please specify or detect change points first.")
    return changepoint_dates


def add_event_window_multi(
        event_df_dict,
        time_col,
        label_col,
        time_delta="1D",
        pre_num=1,
        post_num=1,
        pre_post_num_dict=None):
    """For a given dictionary of events data frames with a time_col and label_col
    it adds shifted events prior and after the given events
    For example if the event data frame includes the row '2019-12-25, Christmas' as a row
    the function will produce dataframes with the events '2019-12-24, Christmas' and '2019-12-26, Christmas' if
    pre_num and post_num are 1 or more.

    Parameters
    ----------
    event_df_dict: `dict` [`str`, `pandas.DataFrame`]
        A dictionary of events data frames
        with each having two columns: ``time_col`` and ``label_col``.
    time_col: `str`
        The column with the timestamp of the events.
        This can be daily but does not have to be.
    label_col : `str`
        The column with labels for the events.
    time_delta : `str`, default "1D"
        The amount of the shift for each unit specified by a string
        e.g. '1D' stands for one day delta
    pre_num : `int`, default 1
        The number of events to be added prior to the given event for each event in df.
    post_num: `int`, default 1
        The number of events to be added after to the given event for each event in df.
    pre_post_num_dict : `dict` [`str`, (`int`, `int`)] or None, default None
        Optionally override ``pre_num`` and ``post_num`` for each key in ``event_df_dict``.
        For example, if ``event_df_dict`` has keys "US" and "India", this parameter
        can be set to ``pre_post_num_dict = {"US": [1, 3], "India": [1, 2]}``,
        denoting that the "US" ``pre_num`` is 1 and ``post_num`` is 3, and "India" ``pre_num`` is 1
        and ``post_num`` is 2. Keys not specified by ``pre_post_num_dict`` use the default given by
        ``pre_num`` and ``post_num``.

    Returns
    -------
    df : `dict` [`str`, `pandas.DataFrame`]
        A dictionary of dataframes for each needed shift. For example if pre_num=2 and post_num=3.
        2 + 3 = 5 data frames will be stored in the return dictionary.
    """
    if pre_post_num_dict is None:
        pre_post_num_dict = {}

    shifted_df_dict = {}

    for event_df_key, event_df in event_df_dict.items():
        if event_df_key in pre_post_num_dict.keys():
            pre_num0 = pre_post_num_dict[event_df_key][0]
            post_num0 = pre_post_num_dict[event_df_key][1]
        else:
            pre_num0 = pre_num
            post_num0 = post_num

        df_dict0 = add_event_window(
            df=event_df,
            time_col=time_col,
            label_col=label_col,
            time_delta=time_delta,
            pre_num=pre_num0,
            post_num=post_num0,
            events_name=event_df_key)

        shifted_df_dict.update(df_dict0)

    return shifted_df_dict


def get_fourier_col_name(k, col_name, function_name="sin", seas_name=None):
    """Returns column name corresponding to a particular fourier term, as returned by fourier_series_fcn

    :param k: int
        fourier term
    :param col_name: str
        column in the dataframe used to generate fourier series
    :param function_name: str
        sin or cos
    :param seas_name: strcols_interact
        appended to new column names added for fourier terms
    :return: str
        column name in DataFrame returned by fourier_series_fcn
    """
    # patsy doesn't allow "." in formula term. Replace "." with "_" rather than quoting "Q()" all fourier terms
    name = f"{function_name}{k:.0f}_{col_name}"
    if seas_name is not None:
        name = f"{name}_{seas_name}"
    return name


def fourier_series_fcn(col_name, period=1.0, order=1, seas_name=None):
    """Generates a function which creates fourier series matrix for a column of an input df
    :param col_name: str
        is the column name in the dataframe which is to be used for
        generating fourier series. It needs to be a continuous variable.
    :param period: float
        the period of the fourier series
    :param order: int
        the order of the fourier series
    :param seas_name: Optional[str]
        appended to new column names added for fourier terms.
        Useful to distinguish multiple fourier
        series on same col_name with different periods.
    :return: callable
        a function which can be applied to any data.frame df
        with a column name being equal to col_name
    """

    def fs_func(df):
        out_df = pd.DataFrame()
        out_cols = []

        if col_name not in df.columns:
            raise ValueError("The data frame does not have the column: " + col_name)
        x = df[col_name]
        x = np.array(x)

        for i in range(order):
            k = i + 1
            sin_col_name = get_fourier_col_name(
                k,
                col_name,
                function_name="sin",
                seas_name=seas_name)
            cos_col_name = get_fourier_col_name(
                k,
                col_name,
                function_name="cos",
                seas_name=seas_name)
            out_cols.append(sin_col_name)
            out_cols.append(cos_col_name)
            omega = 2 * math.pi / period
            u = omega * k * x
            out_df[sin_col_name] = np.sin(u)
            out_df[cos_col_name] = np.cos(u)
        return {"df": out_df, "cols": out_cols}

    return fs_func


def fourier_series_multi_fcn(
        col_names,
        periods=None,
        orders=None,
        seas_names=None):
    """Generates a func which adds multiple fourier series with multiple periods.

    Parameters
    ----------
    col_names : `list` [`str`]
        the column names which are to be used to generate Fourier series.
        Each column can have its own period and order.
    periods:  `list` [`float`] or None
        the periods corresponding to each column given in col_names
    orders : `list` [`int`] or None
        the orders for each of the Fourier series
    seas_names : `list` [`str`] or None
        Appended to the Fourier series name.
        If not provided (None) col_names will be used directly.
    """

    k = len(col_names)
    if periods is None:
        periods = [1.0] * k
    if orders is None:
        orders = [1] * k

    if len(periods) != len(orders):
        raise ValueError("periods and orders must have the same length.")

    def fs_multi_func(df):
        out_df = None
        out_cols = []

        for i in range(k):
            col_name = col_names[i]
            period = periods[i]
            order = orders[i]
            seas_name = None
            if seas_names is not None:
                seas_name = seas_names[i]

            func0 = fourier_series_fcn(
                col_name=col_name,
                period=period,
                order=order,
                seas_name=seas_name)

            res = func0(df)
            fs_df = res["df"]
            fs_cols = res["cols"]
            out_df = pd.concat([out_df, fs_df], axis=1)
            out_cols = out_cols + fs_cols

        return {"df": out_df, "cols": out_cols}

    return fs_multi_func


def signed_pow(x, y):
    """ Takes the absolute value of x and raises it to power of y.
    Then it multiplies the result by sign of x.
    This guarantees this function is non-decreasing.
    This is useful in many contexts e.g. statistical modeling.
    :param x: the base number which can be any real number
    :param y: the power which can be any real number
    :return: returns abs(x) to power of y multiplied by sign of x
    """
    return np.sign(x) * np.power(np.abs(x), y)


def signed_pow_fcn(y):
    return lambda x: signed_pow(x, y)


signed_sqrt = signed_pow_fcn(1 / 2)
signed_sq = signed_pow_fcn(2)


def logistic(x, growth_rate=1.0, capacity=1.0, floor=0.0, inflection_point=0.0):
    """Evaluates the logistic function at x with the specified growth rate,
        capacity, floor, and inflection point.

    :param x: value to evaluate the logistic function
    :type x: float
    :param growth_rate: growth rate
    :type growth_rate: float
    :param capacity: max value (carrying capacity)
    :type capacity: float
    :param floor: min value (lower bound)
    :type floor: float
    :param inflection_point: the t value of the inflection point
    :type inflection_point: float
    :return: value of the logistic function at t
    :rtype: float
    """
    return floor + capacity * expit(growth_rate * (x - inflection_point))


def get_logistic_func(growth_rate=1.0, capacity=1.0, floor=0.0, inflection_point=0.0):
    """Returns a function that evaluates the logistic function at t with the
        specified growth rate, capacity, floor, and inflection point.

        f(x) = floor + capacity / (1 + exp(-growth_rate * (x - inflection_point)))

    :param growth_rate: growth rate
    :type growth_rate: float
    :param capacity: max value (carrying capacity)
    :type capacity: float
    :param floor: min value (lower bound)
    :type floor: float
    :param inflection_point: the t value of the inflection point
    :type inflection_point: float
    :return: the logistic function with specified parameters
    :rtype: callable
    """
    return lambda t: logistic(t, growth_rate, capacity, floor, inflection_point)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Kaixu Yang
"""Utilities to provide summaries of sklearn and statsmodels
regression models.
"""

import numpy as np
import pandas as pd
import statsmodels
import statsmodels.api as sm
from scipy.stats import f
from scipy.stats import norm
from scipy.stats import t
from sklearn.base import clone
from sklearn.linear_model import Ridge
from sklearn.model_selection import train_test_split

from greykite.algo.common.col_name_utils import INTERCEPT
from greykite.algo.common.col_name_utils import simplify_pred_cols


def process_intercept(x, beta, intercept, pred_cols):
    """Processes the intercept term.

    Merges intercept in beta and dummy column in x if they are not there.
    Appends "(Intercept)" to the front of ``pred_cols`` if it is not there.

    Parameters
    ----------
    x : `numpy.array`
        The design matrix.
    beta : `numpy.array`
        The estimated coefficients.
    intercept : `float`
        The estimated intercept.
    pred_cols : `list` [ `str` ]
        The names for predictors.

    Returns
    -------
    x : `numpy.array`
        The design matrix with dummy column.
    beta : `numpy.array`
        The estimated coefficients with intercept in the first position.
    pred_cols : `list` [ `str` ]
        List of names of predictors, with "Intercept" in the first position.
    """
    beta = np.copy(beta.ravel())
    x = np.copy(x)

    # Checks if x has an intercept column
    if not all(x[:, 0] == 1):
        x = np.concatenate([np.ones([x.shape[0], 1]), x], axis=1)
    # Checks beta has intercept column.
    # In this case, the coefficient for intercept column should be zero.
    # Use intercept as the coefficient for beta column
    if x.shape[1] == beta.shape[0]:
        beta[0] += intercept  # combines the intercept term
    # If beta does not have intercept term, concatenate the intercept term.
    elif x.shape[1] == beta.shape[0] + 1:
        beta = np.concatenate([[intercept], beta])
    else:
        raise ValueError("The shape of x and beta do not match. "
                         f"x has shape ({x.shape[0]}, {x.shape[1]}) with the intercept column, "
                         f"but beta has length {len(beta)}.")
    # Processes pred_cols
    if len(pred_cols) == x.shape[1] - 1:
        pred_cols = [INTERCEPT] + pred_cols
    elif len(pred_cols) == x.shape[1] and "Intercept" in pred_cols[0]:
        pass
    else:
        raise ValueError("The length of pred_cols does not match the shape of x. "
                         f"x has shape ({x.shape[0]}, {x.shape[1]}) with the intercept column, "
                         f"but pred_cols has length {len(pred_cols)}.")
    return x, beta, pred_cols


def round_numbers(numbers, sig_digits=4):
    """Rounds numbers to significant figure.

    Rounds numbers to a specified significant figure, and uses string to format it.
    The number will be displayed in scientific mode if there are leading/trailing zeros.
    Displays zero as 0.

    Parameters
    ----------
    numbers : `iterable` [ `float` or `int` ]
        Numbers to be formatted.
    sig_digits : `int`, default 4.
        Significant figure.

    Returns
    -------
    formatted_numbers : `numpy.array`
        Formatted numbers.
    """

    def round_number(number):
        if abs(number) < 1e-40:
            new_num = "0."
        elif ~np.isfinite(number):
            new_num = "nan"
        else:
            num_digit = -int(np.floor(np.log10(abs(number)))) + sig_digits - 1
            new_num = np.round(number, num_digit)
            new_num = ("{:.3e}".format(new_num)
                       if abs(new_num) > 10 ** sig_digits or abs(new_num) < 10 ** (-sig_digits)
                       else str(new_num))
        return new_num

    return np.vectorize(round_number)(numbers)


def create_info_dict_lm(x, y, beta, ml_model, fit_algorithm, pred_cols):
    """Creates a information dictionary for linear model results.

    Only basic information will be created in this function.
    A series of these functions are used in a flow to get all information needed
    in linear model summary. The flow is
    `~greykite.algo.common.model_summary_utils.create_info_dict_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_params_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_df_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_ss_lm`,
    `~greykite.algo.common.model_summary_utils.add_beta_var_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_coef_df_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_significance_lm`.
    For flow implementation, see
    `~greykite.algo.common.model_summary_utils.get_info_dict_lm`.

    Parameters
    ----------
    x : `numpy.array`
        The design matrix.
    y : `numpy.array`
        The response vector.
    beta : `numpy.arrar`
        The estimated coefficients.
    ml_model : `class`
        The trained machine learning model class, see
        `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.
    fit_algorithm : `str`
        The name of fit algorithm, see
        `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.
    pred_cols : `list` [ `str` ]
        A list of predictor names.

    Returns
    -------
    info_dict : `dict`
        A dictionary of basic information with the following keys:

            "x" : `numpy.array`
                The design matrix.
            "y" : `numpy.array`
                The response vector.
            "beta" : `numpy.array`
                The estimated coefficients.
            "ml_model" : `class`
                The trained machine learning model class, see
                `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.
            "fit_algorithm" : `str`
                The name of fit algorithm, see
                `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.
            "pred_cols" : `list` [ `str` ]
                A list of predictor names.
            "degenerate_index" : `numpy.array`
                The indices where the x columns are degenerate.
            "n_sample" : `int`
                Number of observations.
            "n_feature" : `int`
                Number of features.
            "nonzero_index" : `numpy.array`
                Indices of nonzero coefficients.
            "n_feature_nonzero" : `int`
                Number of nonzero coefficients.
            "y_pred" : `numpy.array`
                The predicted values.
            "y_mean" : `float`
                The mean of response.
            "residual" : `numpy.array`
                Residuals.
            "residual_summary" : `numpy.array`
                Five number summary of residuals.
    """
    info_dict = dict()
    # Passes through the given information
    info_dict["x"] = x
    info_dict["y"] = y.ravel()
    info_dict["beta"] = beta.ravel()
    info_dict["ml_model"] = ml_model
    info_dict["fit_algorithm"] = fit_algorithm
    info_dict["pred_cols"] = pred_cols
    # Derives direct information of the input
    info_dict["degenerate_index"] = np.arange(x.shape[1])[x.var(axis=0) == 0]
    info_dict["n_sample"] = x.shape[0]
    info_dict["n_feature"] = x.shape[1]
    info_dict["nonzero_index"] = np.nonzero(beta.ravel())[0]
    info_dict["n_feature_nonzero"] = len(info_dict["nonzero_index"])
    info_dict["y_pred"] = x @ beta
    info_dict["y_mean"] = np.mean(y)
    info_dict["residual"] = info_dict["y"] - info_dict["y_pred"]
    info_dict["residual_summary"] = np.percentile(info_dict["residual"], [0, 25, 50, 75, 100])
    return info_dict


def add_model_params_lm(info_dict):
    """Adds model parameter information to ``info_dict`` for linear models.

    Only model-related information will be added in this function.
    A series of these functions are used in a flow to get all information needed
    in linear model summary. The flow is
    `~greykite.algo.common.model_summary_utils.create_info_dict_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_params_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_df_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_ss_lm`,
    `~greykite.algo.common.model_summary_utils.add_beta_var_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_coef_df_lm`,
    `cadd_model_significance_lm`.
    For flow implementation, see
    `~greykite.algo.common.model_summary_utils.get_info_dict_lm`.

    Parameters
    ----------
    info_dict : `dict`
        The information dictionary returned by
        `~greykite.algo.common.model_summary_utils.create_info_dict_lm`.

    Returns
    -------
    info_dict : `dict`
        The information dictionary with the following keys added:

            "model" : `str`
                The model name.
            "weights" : `numpy.array`, optional
                The weight matrix.
            "family" : `str`, optional
                The distribution family in generalized linear model.
            "link_function" : `str`, optional
                The link function used in generalized linear model.
            "alpha" : `float`, optional
                The regularization parameter in regularized methods.
            "l1_ratio" : `float`, optional
                The l1 norm ratio in Elastic Net methods.
    """
    fit_algorithm = info_dict["fit_algorithm"]
    ml_model = info_dict["ml_model"]
    n_sample = info_dict["n_sample"]
    # Dictionary for model parameters
    valid_linear_fit_algorithms = ["linear", "statsmodels_ols", "statsmodels_wls", "statsmodels_gls", "statsmodels_glm",
                                   "ridge", "lasso", "lars", "lasso_lars", "sgd", "elastic_net"]
    if fit_algorithm in valid_linear_fit_algorithms:
        # Adds special parameters
        if fit_algorithm in ["linear", "statsmodels_ols"]:
            info_dict["model"] = "Ordinary least squares"
        elif fit_algorithm == "statsmodels_wls":
            info_dict["model"] = "Weighted least squares"
            info_dict["weights"] = ml_model.model.weights
        elif fit_algorithm == "statsmodels_gls":
            info_dict["model"] = "Generalized least squares"
            info_dict["weights"] = ml_model.model.sigma
        elif fit_algorithm == "statsmodels_glm":
            info_dict["model"] = "Generalized linear model"
            info_dict["family"] = ml_model.model.family.__class__.__name__
            info_dict["link_function"] = ml_model.model.family.link.__class__.__name__
        elif fit_algorithm in ["ridge", "lasso", "lars", "lasso_lars"]:
            info_dict["model"] = fit_algorithm.capitalize() + " regression"
            info_dict["alpha"] = ml_model.alpha_
        elif fit_algorithm == "elastic_net":
            info_dict["model"] = "Elastic Net regression"
            info_dict["alpha"] = ml_model.alpha_
            info_dict["l1_ratio"] = ml_model.l1_ratio
        elif fit_algorithm == "sgd":
            info_dict["model"] = "Elastic Net regression via SGD"
            info_dict["alpha"] = ml_model.alpha  # does not have underscore, because this is not cv
            if info_dict["ml_model"].penalty == "l1":
                info_dict["l1_ratio"] = 1.0
            elif info_dict["ml_model"].penalty == "l2":
                info_dict["l1_ratio"] = 0.0
            else:
                info_dict["l1_ratio"] = ml_model.l1_ratio

        # Processes parameters
        if "weights" in info_dict:
            if info_dict["weights"] is None:
                weights = np.eye(info_dict["n_sample"])
            else:
                weights = np.array(info_dict["weights"])
                if weights.shape == ():  # checks if weights is a scalar
                    weights = np.diag(np.repeat(weights, n_sample))
                elif weights.shape == (n_sample,) or weights.shape == (n_sample, 1):  # checks if weights is a vector
                    weights = np.diag(weights.ravel())
                elif weights.shape == (n_sample, n_sample):
                    weights = weights
                else:
                    raise ValueError("The shape of weights does not match the design matrix. "
                                     f"The design matrix has length {n_sample}. "
                                     f"The weights has shape {weights.shape}.")
            info_dict["weights"] = weights
    else:
        raise ValueError(f"{fit_algorithm} is not a valid algorithm, it must be in "
                         f"{valid_linear_fit_algorithms}.")
    return info_dict


def add_model_df_lm(info_dict):
    """Adds degrees of freedom of the regression model to ``info_dict`` for linear models.

    The df is defined as the trace of hat matrix. In general, we have
    H=X(X'WX+a*I)^-1X'W and df=trace(H), where X is the design matrix, W is the weight matrix,
    a is the regularization parameter, and I it the identity matrix. For non-weighted methods,
    W is identity matrix, and for non-regularized methods, a is 0.
    For sparse solutions, this is calculated on the nonzero predictors.

    Only degree of freedom information will be added in this function.
    A series of these functions are used in a flow to get all information needed
    in linear model summary. The flow is
    `~greykite.algo.common.model_summary_utils.create_info_dict_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_params_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_df_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_ss_lm`,
    `~greykite.algo.common.model_summary_utils.add_beta_var_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_coef_df_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_significance_lm`.
    For flow implementation, see
    `~greykite.algo.common.model_summary_utils.get_info_dict_lm`.

    Parameters
    ----------
    info_dict : `dict`
        The information dictionary returned by
        `~greykite.algo.common.model_summary_utils.add_model_params_lm`.

    Returns
    -------
    info_dict : `dict`
        The information dictionary with the following keys added:

            "x_nz" : `numpy.array`
                The design matrix with columns corresponding to nonzero estimated coefficients.
            "condition_number" : `float`
                The condition number for sample covariance matrix (weighted, adjusted).
            "xtwx_alphai_inv" : `numpy.array`
                (X'WX+a*I)^-1
            "reg_df" : `float`
                The regression degree of freedom, defined as the trace of hat matrix.
            "df_sse" : `float`
                The degree of freedom of residuals, defined as n_sample - reg_df.
            "df_ssr" : `float`
                The degree of freedom of the regression, defined as reg_df - 1.
            "df_sst" : `int`
                The degree of freedom of total, defined as n_sample - 1.
    """
    x = info_dict["x"]
    n_sample = info_dict["n_sample"]
    w = info_dict.get("weights")
    if w is None:
        w = np.eye(n_sample)
    alpha = info_dict.get("alpha", 0)
    l1_ratio = info_dict.get("l1_ratio", 0)
    # In ElasticNet type models (ElasticNetCV, SGD with penalty == "elasticnet"), the penalty is
    # alpha * l1_ratio * ||beta||_1 + alpha * (1 - l1_ratio) * ||beta||_2^2.
    # For ridge regression, we have l1_ratio = 0, so the l2_norm regularization parameter is alpha.
    # For lasso regression, we have l1_ratio = 1, so the l2_norm regularization parameter is zero.
    alpha = alpha * (1 - l1_ratio)  # if model has l1_ratio, the l2 norm regularization parameter is alpha * (1 - l1_ratio)
    if info_dict["fit_algorithm"] in ["lasso", "lars", "lasso_lars"]:
        alpha = 0  # alpha is specifically for l2 norm regularization in calculating df
    nonzero_idx = info_dict["nonzero_index"]
    x_nz = x[:, nonzero_idx]
    xtwx = x_nz.T @ w @ x_nz
    xtwx_alphai = xtwx + alpha * np.eye(x_nz.shape[1])
    xtwx_alphai_inv = np.linalg.pinv(xtwx_alphai)
    # The hat matrix was x_nz @ xtwx_alphai_inv @ x_nz.T @ w.
    # The regression degrees of freedom is the trace of this hat matrix.
    # This matrix is n x n, which causes memory overflow for large n.
    # Using the trace property trace(ABCD)=trace(CDAB),
    # we compute x_nz.T @ w @ x_nz @ xtwx_alphai_inv,
    # which has dimension p x p.
    trace = np.trace(x_nz.T @ w @ x_nz @ xtwx_alphai_inv)
    info_dict["x_nz"] = x_nz
    info_dict["condition_number"] = np.linalg.cond(xtwx_alphai)
    info_dict["xtwx_alphai_inv"] = xtwx_alphai_inv  # (X'WX+aI)^-1
    info_dict["reg_df"] = trace
    info_dict["df_sse"] = n_sample - info_dict["reg_df"]
    info_dict["df_ssr"] = info_dict["reg_df"] - 1
    info_dict["df_sst"] = n_sample - 1
    return info_dict


def add_model_ss_lm(info_dict):
    """Adds model sum of squared errors to ``info_dict`` for linear models.

    Only sum of squared error information will be added in this function.
    A series of these functions are used in a flow to get all information needed
    in linear model summary. The flow is
    `~greykite.algo.common.model_summary_utils.create_info_dict_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_params_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_df_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_ss_lm`,
    `~greykite.algo.common.model_summary_utils.add_beta_var_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_coef_df_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_significance_lm`.
    For flow implementation, see
    `~greykite.algo.common.model_summary_utils.get_info_dict_lm`.

    Parameters
    ----------
    info_dict : `dict`
        The information dictionary returned by
        `~greykite.algo.common.model_summary_utils.add_model_df_lm`.

    Returns
    -------
    info_dict : `dict`
        The information dictionary with the following keys added:

            "sse" : `float`
                Sum of squared errors from residuals.
            "mse" : `float`
                ``sse`` divided by its degree of freedom.
            "ssr" : `float`
                Sum of squared errors from regression.
            "msr" : `float`
                ``ssr`` divided by its degree of freedom.
            "sst" : `float`
                Sum of squared errors from total.
            "mst" : `float`
                ``sst`` divided by its degree of freedom.
    """
    residual = info_dict["residual"]
    weights = info_dict.get("weights")
    if weights is None:
        weights = np.eye(info_dict["n_sample"])
    y = info_dict["y"]
    y_pred = info_dict["y_pred"]
    y_mean = info_dict["y_mean"]
    info_dict["sse"] = residual.T @ weights @ residual
    info_dict["mse"] = info_dict["sse"] / info_dict["df_sse"]
    info_dict["ssr"] = (y_pred - y_mean).T @ weights @ (y_pred - y_mean)
    info_dict["msr"] = info_dict["ssr"] / info_dict["df_ssr"]
    info_dict["sst"] = (y - y_mean).T @ weights @ (y - y_mean)
    info_dict["mst"] = info_dict["sst"] / info_dict["df_sst"]
    return info_dict


def add_beta_var_lm(info_dict):
    """Adds the covariance matrix for estimated coefficients to `info_dict` for linear models.

    The covariance matrix for estimated coefficients is defined as

    .. code-block:: none

        mse * (X'WX+aI)^-1X'X(X'WX+aI)

    for linear models, and defined as the inverse Fisher information matrix
    divided by ``n_sample`` for generalized linear models.

    Only variance of beta hat information will be added in this function.
    A series of these functions are used in a flow to get all information needed
    in linear model summary. The flow is
    `~greykite.algo.common.model_summary_utils.create_info_dict_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_params_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_df_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_ss_lm`,
    `~greykite.algo.common.model_summary_utils.add_beta_var_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_coef_df_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_significance_lm`.
    For flow implementation, see
    `~greykite.algo.common.model_summary_utils.get_info_dict_lm`.

    Parameters
    ----------
    info_dict : `dict`
        The information dictionary returned by
        `~greykite.algo.common.model_summary_utils.add_model_ss_lm`.

    Returns
    -------
    info_dict : `dict`
        The information dictionary with the following keys added:

            "beta_var_cov" : `numpy.array` or `None`
                The covariance matrix of estimated coefficients. The square root of
                its diagonal elements are standard errors of the estimated coefficients.
                Set as ``None`` for sparse solutions.
    """
    if (info_dict["fit_algorithm"] in ["statsmodels_ols", "statsmodels_wls", "statsmodels_gls", "linear", "ridge"]
            or info_dict["fit_algorithm"] in ["sgd", "elastic_net"] and info_dict["l1_ratio"] == 0):
        xtwx_alphai_inv = info_dict["xtwx_alphai_inv"]
        mse = info_dict["mse"]
        # Variance of estimated coefficients mse * (X'WX+aI)^-1X'X(X'WX+aI)^-1
        x_nz = info_dict["x_nz"]
        info_dict["beta_var_cov"] = xtwx_alphai_inv @ x_nz.T @ x_nz @ xtwx_alphai_inv * mse
    elif info_dict["fit_algorithm"] in ["statsmodels_glm"]:
        # Variance of mle is approximately I(beta_hat)^-1/n, where I is Fisher's information matrix.
        assert isinstance(info_dict["ml_model"], statsmodels.genmod.generalized_linear_model.GLMResultsWrapper)
        info_dict["beta_var_cov"] = info_dict["ml_model"].cov_params()
    else:
        info_dict["beta_var_cov"] = None  # The covariance matrix for beta is not support with sparse solution
    return info_dict


def get_ls_coef_df(info_dict):
    """Gets the coefficients dataframe for least squared models.

    The dataframe includes the estimated values, the standard errors, the t-test
    values, the t-test p-values, the significance code and 95% confidence t-intervals.

    Parameters
    ----------
    info_dict : `dict`
        The information dictionary returned by
        `~greykite.algo.common.model_summary_utils.add_beta_var_lm`.

    Returns
    -------
    info_dict : `dict`
        The information dictionary with the following keys added:

            "coef_summary_df" : `pandas.DataFrame`
                The summary df for estimated coefficients.
            "significance_code_legend" : `str`
                The significance code legend.
    """
    beta = info_dict["beta"]
    std_err = np.sqrt(np.abs(np.diag(info_dict["beta_var_cov"])))
    if len(std_err) != len(beta):
        full_std_err = np.full(beta.shape, np.inf)
        full_std_err[info_dict["nonzero_index"]] = std_err
        full_std_err[full_std_err == 0] = np.inf  # These are the degenerate columns.
        std_err = full_std_err
    # Gets t-values
    t_values = beta / std_err

    # Gets p-values and significance codes
    p_values = 2 * t.cdf(-np.abs(t_values), info_dict["df_sse"]).ravel()
    sig_codes = ["***" if p < 0.001 else
                 "**" if p < 0.01 else
                 "*" if p < 0.05 else
                 "." if p < 0.1 else
                 " " for p in p_values]
    # Gets 95% confidence intervals
    t_975 = t.ppf(0.975, info_dict["df_sse"])
    ci_lb = beta - t_975 * std_err
    ci_ub = beta + t_975 * std_err
    ci = list(zip(ci_lb, ci_ub))
    # Makes full df
    info_dict["coef_summary_df"] = pd.DataFrame({
        "Pred_col": info_dict["pred_cols"],
        "Estimate": beta,
        "Std. Err": std_err,
        "t value": t_values,
        "Pr(>|t|)": p_values,
        "sig. code": sig_codes,
        "95%CI": ci
    })
    info_dict["significance_code_legend"] = "0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1"
    return info_dict


def get_glm_coef_df(info_dict):
    """Gets the coefficients dataframe for generalized linear models.

    The dataframe includes the estimated values, the standard errors, the Z-test
    values, the Z-test p-values, the significance code and 95% confidence Z-intervals.
    The tests and intervals are based on the asymptotic normality of MLE.

    Parameters
    ----------
    info_dict : `dict`
        The information dictionary returned by
        `~greykite.algo.common.model_summary_utils.add_beta_var_lm`.

    Returns
    -------
    info_dict : `dict`
        The information dictionary with the following keys added:

            "coef_summary_df" : `pandas.DataFrame`
                The summary df for estimated coefficients.
            "significance_code_legend" : `str`
                The significance code legend.
    """
    beta = info_dict["beta"]
    std_err = np.sqrt(np.abs(np.diag(info_dict["beta_var_cov"])))
    # Gets z-values
    z_values = beta / std_err
    # Gets p-values and significance codes
    p_values = 2 * norm.cdf(-np.abs(z_values)).ravel()
    sig_codes = ["***" if p < 0.001 else
                 "**" if p < 0.01 else
                 "*" if p < 0.05 else
                 "." if p < 0.1 else
                 " " for p in p_values]
    # Gets 95% confidence intervals
    z_975 = norm.ppf(0.975)
    ci_lb = beta - z_975 * std_err
    ci_ub = beta + z_975 * std_err
    ci = list(zip(ci_lb, ci_ub))
    # Makes full df
    info_dict["coef_summary_df"] = pd.DataFrame({
        "Pred_col": info_dict["pred_cols"],
        "Estimate": beta,
        "Std. Err": std_err,
        "z value": z_values,
        "Pr(>|Z|)": p_values,
        "sig. code": sig_codes,
        "95%CI": ci
    })
    info_dict["significance_code_legend"] = "0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1"
    return info_dict


class Bootstrapper(object):
    """Bootstrap generator, an iterable object.

    The generator for bootstrap samples. Iterating along the object will
    give the bootstrap sample indices.

    Attributes
    ----------
    sample_size : `int`
        The sample size of data.
    bootstrap_size : `int`
        The bootstrap sample size to be randomly drawn.
    num_bootstrap : `int`
        The number of bootstrap samples to be drawn.
    sample_num : `int`
        The index of number of samples generated.
    """

    def __init__(self, sample_size, bootstrap_size, num_bootstrap):
        self.sample_size = sample_size
        self.bootstrap_size = bootstrap_size
        self.num_bootstrap = num_bootstrap
        self.sample_num = 0

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        if self.sample_num < self.num_bootstrap:
            sample_index = np.random.choice(
                a=self.sample_size,
                size=self.bootstrap_size,
                replace=True)
            self.sample_num += 1
            return sample_index
        else:
            raise StopIteration


def get_ridge_summary_df_by_bootstrap(x, y, beta, alpha, pred_cols):
    """Gets the ridge regression coefficients p-values and confidence intervals.

    The tests and confidence intervals are bootstrap p-values and confidence intervals.

    Parameters
    ----------
    x : `numpy.array`
        The design matrix.
    y : `numpy.array`
        The response vector.
    beta : `numpy.array`
        The estimated coefficients.
    alpha : `float`
        The regularization parameters.
    pred_cols : `list` [ `str` ]
        The list of predictor names.

    Returns
    -------
    coef_summary_df : `pandas.DataFrame`
        The coefficients summary dataframe.
    significance_code_legend : `str`
        The significance code legend.
    """
    num_bootstrap = 500
    bootstrap_idx = Bootstrapper(
        sample_size=x.shape[0],
        bootstrap_size=x.shape[0],
        num_bootstrap=num_bootstrap)
    beta_estimates = None
    intercepts = []
    for idx in bootstrap_idx:
        x_b = x[idx, :]
        y_b = y[idx]
        model = Ridge(alpha=alpha, fit_intercept=True).fit(x_b, y_b)  # Fits intercept separately, does not penalize intercept
        intercepts.append(model.intercept_)
        # ``beta_estimates`` stores the bootstrapped beta: col = bootstrap indices; rows=betas
        if beta_estimates is None:
            beta_estimates = model.coef_.reshape(-1, 1)
        else:
            beta_estimates = np.concatenate([beta_estimates, model.coef_.reshape(-1, 1)], axis=1)
    beta_estimates[0] += np.array(intercepts).reshape(beta_estimates[0].shape)  # adds intercepts
    # Gets coefficient standard errors
    std_err = beta_estimates.std(axis=1)

    # Reference: An introduction to Bootstrap, Efron 1993

    # Gets p-values and significance codes
    # For an estimated coefficient, its p-value is the proportion of bootstrap estimates whose distances
    # to the bootstrap distribution mean are at least the absolute value of the estimated coefficient.
    p_values = np.apply_along_axis(
        func1d=lambda row: len(row[:-1][abs(row[:-1] - row[:-1].mean()) >= abs(row[-1])]),
        axis=1,
        arr=np.concatenate([beta_estimates, beta.reshape(-1, 1)], axis=1)) / num_bootstrap
    sig_codes = ["***" if p < 0.001 else
                 "**" if p < 0.01 else
                 "*" if p < 0.05 else
                 "." if p < 0.1 else
                 " " for p in p_values]

    # Gets 95% confidence intervals
    ci_lb = np.percentile(beta_estimates, 2.5, axis=1)
    ci_ub = np.percentile(beta_estimates, 97.5, axis=1)
    ci = [[lb, ub] for lb, ub in zip(ci_lb, ci_ub)]

    # Makes full df
    coef_summary_df = pd.DataFrame({
        "Pred_col": pred_cols,
        "Estimate": beta,
        "Std. Err": std_err,
        "Pr(>)_boot": p_values,
        "sig. code": sig_codes,
        "95%CI": ci
    })
    return coef_summary_df, "0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1"


def get_ridge_coef_df(info_dict):
    """Gets the coefficients dataframe for ridge regression models.

    The dataframe includes the estimated values, the standard errors,
    the significance test p-values, the significance code and 95% confidence intervals.
    The tests and confidence intervals are based on bootstrap techniques.

    Parameters
    ----------
    info_dict : `dict`
        The information dictionary returned by
        `~greykite.algo.common.model_summary_utils.add_beta_var_lm`.

    Returns
    -------
    info_dict : `dict`
        The information dictionary with the following keys added:

            "coef_summary_df" : `pandas.DataFrame`
                The summary df for estimated coefficients.
            "significance_code_legend" : `str`
                The significance code legend.
    """
    info_dict["coef_summary_df"], info_dict["significance_code_legend"] = get_ridge_summary_df_by_bootstrap(
        x=info_dict["x"],
        y=info_dict["y"],
        beta=info_dict["beta"],
        alpha=info_dict["alpha"],
        pred_cols=info_dict["pred_cols"]
    )
    return info_dict


def get_lasso_coef_by_single_sample_split(info_dict):
    """Gets the lasso coefficients' and metrics on a single sample-splitting.

    Randomly splits the data into two parts of equal size. Performs the same
    lasso regression method as ``info_dict["ml_model"]`` to train on
    one part to identify the significant features, then trains OLS on the other
    part with the selected features. Returns the estimated coefficients and the
    standard errors.

    Parameters
    ----------
    info_dict : `dict`
        The information dictionary returned by
        `~greykite.algo.common.model_summary_utils.add_beta_var_lm`.

    Returns
    -------
    beta : `numpy.array`
        The estimated coefficients. For those features that are selected with
        the first half data, the coefficients are estimated with OLS on the other
        half data. For those features that have coefficients zero with the first
        half data, their coefficients are zero.
    std_err : `numpy.array`
        The standard errors for the estimated coefficients. For zero coefficients,
        the standard errors are set to 1 for the convenience of future calculations,
        they are not the real standard errors.
    df_mse : `float`
        The degree of freedom of the error.
    """
    # Gets indices for the split
    x_0, x_1, y_0, y_1 = train_test_split(
        info_dict["x"],
        info_dict["y"],
        test_size=0.5,
        shuffle=True)
    # Performs lasso on x_0 and y_0
    model = clone(info_dict["ml_model"]).fit(x_0, y_0)
    # Gets nonzero features' index
    coef = model.coef_
    coef[0] = model.intercept_
    nonzero_index = (model.coef_ != 0)
    # Trains regular linear regression on the nonzero features with x_1 and y_1
    x_2 = x_1[:, nonzero_index]  # nonzero features only
    nonzero_beta = sm.OLS(y_1, x_2).fit().params
    beta = np.zeros(len(coef))
    beta[nonzero_index] = nonzero_beta
    df_mse = x_2.shape[0] - np.trace(x_2 @ np.linalg.pinv(x_2.T @ x_2) @ x_2.T)
    y_pred = x_2 @ nonzero_beta
    residuals = y_1 - y_pred
    mse = np.sum(residuals ** 2) / df_mse
    covariance = mse * np.linalg.pinv(x_2.T @ x_2)
    standard_err = np.sqrt(np.diag(covariance))  # nonzero standard errors
    # The standard errors for the zero betas are not one, but since they should not be
    # used and to avoid warnings for possible divisions, we set them to be one.
    # Remember these are not the real standard errors for the zero beta.
    std_err = np.ones(len(beta))
    std_err[nonzero_index] = standard_err
    return beta, std_err, df_mse


def get_lasso_coef_df_by_multi_sample_split(info_dict):
    """Gets the lasso coefficients' p-values and confidence intervals by multi sample-splitting.

    Repeats the single sample-splitting in
    `~greykite.algo.common.model_summary_utils.get_lasso_coef_by_single_sample_split`
    for multiple times, and aggregate the p-values and confidence intervals according to
    "High-Dimensional Inference: Confidence Intervals, p-Values and R-Software hdi"
    by Dezeure, Buhlmann, Meier and Meinshausen

    Parameters
    ----------
    info_dict : `dict`
        The information dictionary returned by
        `~greykite.algo.common.model_summary_utils.add_beta_var_lm`.
        Note that the "ml_model" in ``info_dict`` should be a lasso method
        that supports cloning.

    Returns
    -------
    coef_summary_df : `pandas.DataFrame`
        The coefficients summary dataframe.
    significance_code_legend : `str`
        The significance code legend.
    """
    # Keeping ``num_splits`` as 100 is a good choice.
    # Less number of splits make some values inaccurate (prob_nonzero, p_value, etc.).
    # Greater number of splits takes longer to run (each split runs a LassoCV).
    num_splits = 100
    beta_estimates = None
    standard_errs = None
    df_mses = np.array([])
    # beta_estimates includes the betas estimated by multi sample-splitting.
    # Each row is a coefficient, and each column is an estimation from a sample-splitting.
    for i in range(num_splits):
        beta, standard_err, df_mse = get_lasso_coef_by_single_sample_split(info_dict)
        if beta_estimates is None:
            beta_estimates = beta.reshape(-1, 1)
        else:
            beta_estimates = np.concatenate([beta_estimates, beta.reshape(-1, 1)], axis=1)
        if standard_errs is None:
            standard_errs = standard_err.reshape(-1, 1)
        else:
            standard_errs = np.concatenate([standard_errs, standard_err.reshape(-1, 1)], axis=1)
        df_mses = np.append(df_mses, df_mse)
    # Calculates nonzero probability for each feature coefficient
    beta_nonzero_prob = np.count_nonzero(beta_estimates, axis=1) / beta_estimates.shape[1]

    # Calculates individual p-values
    beta_estimates_p_values = np.where(beta_estimates != 0,
                                       2 * t.cdf(-np.abs(beta_estimates / standard_errs), df_mses),
                                       1)  # p-value is 1.0 for zero coefficients
    # Aggregates indivudual p-values
    # Reference: P-Values for High-dimensional Regression
    # by Meinshausen, Meier, Buhlmann
    # Link: https://arxiv.org/pdf/0811.2177.pdf
    gamma_min = 0.05  # a lower bound for the p-value percentile, typically 0.05
    gammas = np.arange(gamma_min, 1, 0.01)
    qjs = None
    # For each percentile gamma, the qj is defined as the gamma_th percentile of p-values divided by gamma
    # Set to 1 if the value exceeds 1.
    for gamma in gammas:
        qj = np.percentile(beta_estimates_p_values, q=gamma * 100, axis=1) / gamma
        qj = np.where(qj < 1, qj, 1)  # set to 1 if the value exceeds 1
        if qjs is None:
            qjs = qj.reshape(-1, 1)
        else:
            qjs = np.concatenate([qjs, qj.reshape(-1, 1)], axis=1)
    p_values = np.min(qjs, axis=1) * (1 - np.log(gamma_min))
    p_values = np.where(p_values < 1, p_values, 1)
    sig_codes = ["***" if p < 0.001 else
                 "**" if p < 0.01 else
                 "*" if p < 0.05 else
                 "." if p < 0.1 else
                 " " for p in p_values]

    # Calculates p-value rankings, plus one because we want ranking to start with 1
    p_value_rankings = (beta_estimates_p_values.argsort(axis=1) + 1) / beta_estimates_p_values.shape[1]
    # Calculates the confidence intervals
    # Reference: High-Dimensional Inference: Confidence Intervals, p-Values and R-Software hdi
    # by Dezeure, Buhlmann, Meier and Meinshausen
    alpha = 0.05
    # Divides by 2 because we need half of the area on each side
    quantile = 1 - alpha * p_value_rankings / (1 - np.log(gamma_min)) / 2
    # Raveled dfs are applied along columns
    ordinary_confidence_interval_lb = beta_estimates - t.ppf(quantile, df_mses.ravel()) * standard_errs
    ordinary_confidence_interval_ub = beta_estimates + t.ppf(quantile, df_mses.ravel()) * standard_errs

    confidence_intervals = []
    for i in range(beta_estimates.shape[0]):
        if (beta_estimates[i] == 0).all():
            confidence_intervals.append([0., 0.])
        else:
            check_index = (p_value_rankings[i] > gamma_min) & (beta_estimates[i] != 0)
            if len(check_index) == 0:
                confidence_intervals.append([0., 0.])
            else:
                lb = np.min(ordinary_confidence_interval_lb[i, check_index])
                ub = np.max(ordinary_confidence_interval_ub[i, check_index])
                confidence_intervals.append([lb, ub])

    # Makes summary df
    coef_summary_df = pd.DataFrame({
        "Pred_col": info_dict["pred_cols"],
        "Estimate": info_dict["beta"],
        "Pr(>)_split": p_values,
        "sig. code": sig_codes,
        "95%CI": confidence_intervals,
        "Prob_nonzero": beta_nonzero_prob
    })
    return coef_summary_df, "0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1"


def get_lasso_coef_df(info_dict):
    """Gets the coefficients dataframe for lasso regression models.

    The dataframe includes the estimated values, the probability a coefficient is nonzero,
    the significance test p-values, the significance code and 95% confidence intervals.
    The tests and confidence intervals are based on multi sample-splitting techniques.

    Parameters
    ----------
    info_dict : `dict`
        The information dictionary returned by
        `~greykite.algo.common.model_summary_utils.add_beta_var_lm`.

    Returns
    -------
    info_dict : `dict`
        The information dictionary with the following keys added:

            "coef_summary_df" : `pandas.DataFrame`
                The summary df for estimated coefficients.
            "significance_code_legend" : `str`
                The significance code legend.
    """
    info_dict["coef_summary_df"], info_dict["significance_code_legend"] = get_lasso_coef_df_by_multi_sample_split(info_dict)
    return info_dict


def get_elasticnet_coef_df(info_dict):
    """Gets the coefficients dataframe for Elastic Net regression models.

    Currently only returns the predictor names and their corresponding
    estimated coefficients.

    Parameters
    ----------
    info_dict : `dict`
        The information dictionary returned by
        `~greykite.algo.common.model_summary_utils.add_beta_var_lm`.

    Returns
    -------
    info_dict : `dict`
        The information dictionary with the following keys added:

            "coef_summary_df" : `pandas.DataFrame`
                The summary df for estimated coefficients.
    """
    info_dict["coef_summary_df"] = pd.DataFrame({
        "Pred_col": info_dict["pred_cols"],
        "Estimate": info_dict["beta"]
    })
    return info_dict


def add_model_coef_df_lm(info_dict):
    """Adds the tests and confidence intervals for estimated coefficients
       to `info_dict` for linear models.

    Only tests and confidence intervals information will be added in this function.
    A series of these functions are used in a flow to get all information needed
    in linear model summary. The flow is
    `~greykite.algo.common.model_summary_utils.create_info_dict_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_params_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_df_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_ss_lm`,
    `~greykite.algo.common.model_summary_utils.add_beta_var_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_coef_df_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_significance_lm`.
    For flow implementation, see
    `~greykite.algo.common.model_summary_utils.get_info_dict_lm`.

    Parameters
    ----------
    info_dict : `dict`
        The information dictionary returned by
        `~greykite.algo.common.model_summary_utils.add_beta_var_lm`.

    Returns
    -------
    info_dict : `dict`
        The information dictionary with the following keys added:

            "coef_summary_df" : `pandas.DataFrame`
                The summary df for estimated coefficients.
            "significance_code_legend" : `str`, optional
                The significance code legend.
    """
    if info_dict["fit_algorithm"] in ["statsmodels_ols", "statsmodels_wls", "statsmodels_gls", "linear"]:
        info_dict = get_ls_coef_df(info_dict)
    elif info_dict["fit_algorithm"] in ["statsmodels_glm"]:
        info_dict = get_glm_coef_df(info_dict)
    elif (info_dict["fit_algorithm"] in ["ridge"]
          or (info_dict["fit_algorithm"] in ["sgd", "elastic_net"] and info_dict["l1_ratio"] == 0)):
        info_dict = get_ridge_coef_df(info_dict)
    elif (info_dict["fit_algorithm"] in ["lasso", "lars", "lasso_lars"]
          or (info_dict["fit_algorithm"] in ["elastic_net"] and info_dict["l1_ratio"] == 1)):
        info_dict = get_lasso_coef_df(info_dict)
    elif info_dict["fit_algorithm"] in ["sgd", "elastic_net"]:
        info_dict = get_elasticnet_coef_df(info_dict)
    return info_dict


def add_model_significance_lm(info_dict):
    """Adds model significance metrics to `info_dict` for linear models.

    Only model significance metrics information will be added in this function.
    A series of these functions are used in a flow to get all information needed
    in linear model summary. The flow is
    `~greykite.algo.common.model_summary_utils.create_info_dict_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_params_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_df_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_ss_lm`,
    `~greykite.algo.common.model_summary_utils.add_beta_var_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_coef_df_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_significance_lm`.
    For flow implementation, see
    `~greykite.algo.common.model_summary_utils.get_info_dict_lm`.

    Parameters
    ----------
    info_dict : `dict`
        The information dictionary returned by
        `~greykite.algo.common.model_summary_utils.add_model_coef_df_lm`.

    Returns
    -------
    info_dict : `dict`
        The information dictionary with the following keys added:

            "f_value" : `float`
                The F-ratio for model significance.
                Defined as ``msr / mse``
            "f_p_value" : `float`
                The p-value of F-ratio using degrees of freedom
                ``df_ssr`` and ``df_sse``.
            "r2" : `float`
                The coefficient of determination.
                Defined as ``ssr / sst``
            "r2_adj" : `float`
                The adjusted coefficient of determination.
                Defined as ``msr / mst``
            "aic" : `float`
                The AIC of model, with the constants depending only on
                ``n_sample`` removed.
            "bic" : `float`
                The BIC of model, with the constants depending only on
                ``n_sample`` removed.
    """
    # Gets F-value and its p-value
    info_dict["f_value"] = info_dict["msr"] / info_dict["mse"]
    info_dict["f_p_value"] = 1 - f.cdf(info_dict["f_value"], info_dict["df_ssr"], info_dict["df_sse"])
    # Gets R^2 and adjusted R^2
    info_dict["r2"] = 1 - info_dict["sse"] / info_dict["sst"]
    info_dict["r2_adj"] = 1 - info_dict["mse"] / info_dict["mst"]
    # Gets AIC and BIC
    if info_dict["fit_algorithm"] == "statsmodels_glm":
        info_dict["aic"] = info_dict["ml_model"].aic
        info_dict["bic"] = info_dict["ml_model"].bic
    else:
        info_dict["aic"] = 2 * info_dict["reg_df"] + info_dict["n_sample"] * np.log(info_dict["sse"])  # Ignores the constants
        info_dict["bic"] = np.log(info_dict["n_sample"]) * info_dict["reg_df"] + info_dict["n_sample"] * np.log(info_dict["sse"])  # Ignores the constants
    return info_dict


def get_info_dict_lm(x, y, beta, ml_model, fit_algorithm, pred_cols):
    """Get the ``info_dict`` dictionary for linear models.

    A series of functions are used in a flow to get all information needed
    in linear model summary. The flow is
    `~greykite.algo.common.model_summary_utils.create_info_dict_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_params_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_df_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_ss_lm`,
    `~greykite.algo.common.model_summary_utils.add_beta_var_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_coef_df_lm`,
    `~greykite.algo.common.model_summary_utils.add_model_significance_lm`.

    Parameters
    ----------
    x : `numpy.array`
        The design matrix.
    y : `numpy.array`
        The response vector.
    beta : `numpy.array`
        The estimated coefficients.
    ml_model : `class`
        The trained machine learning model class, see
        `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.
    fit_algorithm : `str`
        The name of fit algorithm, see
        `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.
    pred_cols : `list` [ `str` ]
        A list of predictor names.

    Returns
    -------
    info_dict : `dict`
        The dictionary of linear model summary information with the following keys:

            "x" : `numpy.array`
                The design matrix.
            "y" : `numpy.array`
                The response vector.
            "beta" : `numpy.array`
                The estimated coefficients.
            "ml_model" : `class`
                The trained machine learning model class, see
                `~greykite.algo.common.ml_models.fit_model_via_design_matrix`
            "fit_algorithm" : `str`
                The name of fit algorithm, see
                `~greykite.algo.common.ml_models.fit_model_via_design_matrix`
            "pred_cols" : `list` [ `str` ]
                A list of predictor names.
            "degenerate_index" : `numpy.array`
                The indices where the x columns are degenerate.
            "n_sample" : `int`
                Number of observations.
            "n_feature" : `int`
                Number of features.
            "nonzero_index" : `numpy.array`
                Indices of nonzero coefficients.
            "n_feature_nonzero" : `int`
                Number of nonzero coefficients.
            "y_pred" : `numpy.array`
                The predicted values.
            "y_mean" : `float`
                The mean of response.
            "residual" : `numpy.array`
                Residuals.
            "residual_summary" : `numpy.array`
                Five number summary of residuals.
            "model" : `str`
                The model name.
            "weights" : `numpy.array`, optional
                The weight matrix.
            "family" : `str`, optional
                The distribution family in generalized linear model.
            "link_function" : `str`, optional
                The link function used in generalized linear model.
            "alpha" : `float`, optional
                The regularization parameter in regularized methods.
            "l1_ratio" : `float`, optional
                The l1 norm ratio in Elastic Net methods.
            "x_nz" : `numpy.array`
                The design matrix with columns corresponding to nonzero estimated coefficients.
            "condition_number" : `float`
                The condition number for sample covariance matrix (weighted, adjusted).
            "xtwx_alphai_inv" : `numpy.array`
                (X'WX+a*I)^-1
            "reg_df" : `float`
                The regression degree of freedom, defined as the trace of hat matrix.
            "df_sse" : `float`
                The degree of freedom of residuals, defined as n_sample - reg_df.
            "df_ssr" : `float`
                The degree of freedom of the regression, defined as reg_df - 1.
            "df_sst" : `int`
                The degree of freedom of total, defined as n_sample - 1.
            "sse" : `float`
                Sum of squared errors from residuals.
            "mse" : `float`
                ``sse`` divided by its degree of freedom.
            "ssr" : `float`
                Sum of squared errors from regression.
            "msr" : `float`
                ``ssr`` divided by its degree of freedom.
            "sst" : `float`
                Sum of squared errors from total.
            "mst" : `float`
                ``sst`` divided by its degree of freedom.
            "beta_var_cov" : `numpy.array`
                The covariance matrix of estimated coefficients. The squared root of
                its diagonal elements are standard errors of the estimated coefficients.
                Set as ``None`` for sparse solutions.
            "coef_summary_df" : `pandas.DataFrame`
                The summary df for estimated coefficients.
            "significance_code_legend" : `str`, optional
                The significance code legend.
            "f_value" : `float`
                The F-ratio for model significance.
                Defined as ``msr / mse``
            "f_p_value" : `float`
                The p-value of F-ratio using degrees of freedom
                ``df_ssr`` and ``df_sse``.
            "r2" : `float`
                The coefficient of determination.
                Defined as ``ssr / sst``
            "r2_adj" : `float`
                The adjusted coefficient of determination.
                Defined as ``msr / mst``
            "aic" : `float`
                The AIC of model, with the constants depending only on
                ``n_sample`` removed.
            "bic" : `float`
                The BIC of model, with the constants depending only on
                ``n_sample`` removed.
            "model_type" : `str`
                Equals "lm".
    """
    info_dict = create_info_dict_lm(x, y, beta, ml_model, fit_algorithm, pred_cols)
    info_dict = add_model_params_lm(info_dict)
    info_dict = add_model_df_lm(info_dict)
    info_dict = add_model_ss_lm(info_dict)
    info_dict = add_beta_var_lm(info_dict)
    info_dict = add_model_coef_df_lm(info_dict)
    info_dict = add_model_significance_lm(info_dict)
    info_dict["model_type"] = "lm"
    return info_dict


def create_info_dict_tree(x, y, ml_model, fit_algorithm, pred_cols):
    """Creates a information dictionary for tree model results.

    Only basic information will be created in this function.
    A series of these functions are used in a flow to get all information needed
    in tree model summary. The flow is
    `~greykite.algo.common.model_summary_utils.create_info_dict_tree`,
    `~greykite.algo.common.model_summary_utils.add_model_params_tree`,
    `~greykite.algo.common.model_summary_utils.add_model_coef_df_tree`.
    For flow implementation, see
    `~greykite.algo.common.model_summary_utils.get_info_dict_tree`.

    Parameters
    ----------
    x : `numpy.array`
        The design matrix.
    y : `numpy.array`
        The response vector.
    ml_model : `class`
        The trained machine learning model class, see
        `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.
    fit_algorithm : `str`
        The name of fit algorithm, see
        `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.
    pred_cols : `list` [ `str` ]
        A list of predictor names.

    Returns
    -------
    info_dict : `dict`
        A dictionary of basic information with the following keys:

            "x" : `numpy.array`
                The design matrix.
            "y" : `numpy.array`
                The response vector.
            "ml_model" : `class`
                The trained machine learning model class, see
                `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.
            "fit_algorithm" : `str`
                The name of fit algorithm, see
                `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.
            "pred_cols" : `list` [ `str` ]
                A list of predictor names.
            "degenerate_index" : `numpy.array`
                The indices where the x columns are degenerate.
            "n_sample" : `int`
                Number of observations.
            "n_feature" : `int`
                Number of features.
            "y_pred" : `numpy.array`
                The predicted values.
            "y_mean" : `float`
                The mean of response.
            "residual" : `numpy.array`
                Residuals.
            "residual_summary" : `numpy.array`
                Five number summary of residuals.
    """
    info_dict = dict()
    # Passes through the given information
    info_dict["x"] = x
    info_dict["y"] = y
    info_dict["ml_model"] = ml_model
    info_dict["fit_algorithm"] = fit_algorithm
    info_dict["pred_cols"] = pred_cols
    # Derives simple information
    info_dict["n_sample"] = x.shape[0]
    info_dict["n_feature"] = x.shape[1]
    info_dict["y_pred"] = ml_model.predict(x)
    info_dict["y_mean"] = np.mean(y)
    info_dict["residual"] = info_dict["y"] - info_dict["y_pred"]
    info_dict["residual_summary"] = np.percentile(info_dict["residual"], [0, 25, 50, 75, 100])
    return info_dict


def add_model_params_tree(info_dict):
    """Adds model parameters to `info_dict` for tree models.

    Only model-related parameter information will be added in this function.
    A series of these functions are used in a flow to get all information needed
    in tree model summary. The flow is
    `~greykite.algo.common.model_summary_utils.create_info_dict_tree`,
    `~greykite.algo.common.model_summary_utils.add_model_params_tree`,
    `~greykite.algo.common.model_summary_utils.add_model_coef_df_tree`.
    For flow implementation, see
    `~greykite.algo.common.model_summary_utils.get_info_dict_tree`.

    Parameters
    ----------
    info_dict : `dict`
        The information dictionary returned by
        `~greykite.algo.common.model_summary_utils.create_info_dict_tree`.

    Returns
    -------
    info_dict : `dict`
        The information dictionary with the following keys added:

            "model" : `str`
                The model name.
            "num_tree" : `int`
                Number of trees used.
            "criterion" : `str`
                The criterion to be minimized.
            "max_depth" : `int` or `None`
                The maximal tree depth.
            "subsample" : `float`, `int` or `None`
                The subsampling proportion or a whole number to sample.
            "max_features" : `int` or `None`
                The maximal number of features to be used in a single split.
    """
    fit_algorithm = info_dict["fit_algorithm"]
    ml_model = info_dict["ml_model"]
    valid_tree_fit_algorithms = ["rf", "gradient_boosting"]
    if fit_algorithm in valid_tree_fit_algorithms:
        if fit_algorithm == "gradient_boosting":
            info_dict["model"] = "Gradient Boosting"
            info_dict["num_tree"] = ml_model.n_estimators_
            info_dict["criterion"] = ml_model.criterion
            info_dict["max_depth"] = ml_model.max_depth
            info_dict["subsample"] = ml_model.subsample
            info_dict["max_features"] = ml_model.max_features_
        elif fit_algorithm == "rf":
            info_dict["model"] = "Random Forest"
            info_dict["num_tree"] = ml_model.n_estimators
            info_dict["criterion"] = ml_model.criterion
            info_dict["max_depth"] = ml_model.max_depth
            info_dict["subsample"] = ml_model.max_samples
            info_dict["max_features"] = ml_model.max_features
    else:
        raise ValueError(f"{fit_algorithm} is not a valid algorithm, it must be in "
                         f"{valid_tree_fit_algorithms}.")
    return info_dict


def add_model_coef_df_tree(info_dict):
    """Adds coefficient summary df to `info_dict` for tree models.

    Only coefficient summary information will be added in this function.
    A series of these functions are used in a flow to get all information needed
    in tree model summary. The flow is
    `~greykite.algo.common.model_summary_utils.create_info_dict_tree`,
    `~greykite.algo.common.model_summary_utils.add_model_params_tree`,
    `~greykite.algo.common.model_summary_utils.add_model_coef_df_tree`.
    For flow implementation, see
    `~greykite.algo.common.model_summary_utils.get_info_dict_tree`.

    Parameters
    ----------
    info_dict : `dict`
        The information dictionary returned by
        `~greykite.algo.common.model_summary_utils.add_model_params_tree`.

    Returns
    -------
    info_dict : `dict`
        The information dictionary with the following keys added:

            "coef_summary_df" : `pandas.DataFrame`
                The model summary df with the following columns:

                    "Pred_col" : `str`
                        The predictor names.
                    "Feature importance" : `float`
                        The impurity-based feature importance.
                    "Importance rank" : `int`
                        The rank of feature importance.
    """
    feature_importance = info_dict["ml_model"].feature_importances_
    # Gets ranks
    temp = feature_importance.argsort()
    ranks = np.empty_like(temp)
    ranks[temp] = np.arange(len(feature_importance))
    coef_df = pd.DataFrame({
        "Pred_col": info_dict["pred_cols"],
        "Feature importance": feature_importance,
        "Importance rank": len(feature_importance) - ranks  # Reverse order
    })
    info_dict["coef_summary_df"] = coef_df
    return info_dict


def get_info_dict_tree(x, y, ml_model, fit_algorithm, pred_cols):
    """Get the ``info_dict`` dictionary for tree models.

    A series of functions are used in a flow to get all information needed
    in tree model summary. The flow is
    `~greykite.algo.common.model_summary_utils.create_info_dict_tree`,
    `~greykite.algo.common.model_summary_utils.add_model_params_tree`,
    `~greykite.algo.common.model_summary_utils.add_model_coef_df_tree`.

    Parameters
    ----------
    x : `numpy.array`
        The design matrix.
    y : `numpy.array`
        The response vector.
    ml_model : `class`
        The trained machine learning model class, see
        `~greykite.algo.common.ml_models.fit_model_via_design_matrix`
    fit_algorithm : `str`
        The name of fit algorithm, see
        `~greykite.algo.common.ml_models.fit_model_via_design_matrix`
    pred_cols : `list` [ `str` ]
        A list of predictor names.

    Returns
    -------
    info_dict : `dict`
        The dictionary of tree model summary information with the following keys:

            "x" : `numpy.array`
                The design matrix.
            "y" : `numpy.array`
                The response vector.
            "ml_model" : `class`
                The trained machine learning model class, see
                `~greykite.algo.common.ml_models.fit_model_via_design_matrix`
            "fit_algorithm" : `str`
                The name of fit algorithm, see
                `~greykite.algo.common.ml_models.fit_model_via_design_matrix`
            "pred_cols" : `list` [ `str` ]
                A list of predictor names.
            "degenerate_index" : `numpy.array`
                The indices where the x columns are degenerate.
            "n_sample" : `int`
                Number of observations.
            "n_feature" : `int`
                Number of features.
            "y_pred" : `numpy.array`
                The predicted values.
            "y_mean" : `float`
                The mean of response.
            "residual" : `numpy.array`
                Residuals.
            "residual_summary" : `numpy.array`
                Five number summary of residuals.
            "model" : `str`
                The model name.
            "num_tree" : `int`
                Number of trees used.
            "criterion" : `str`
                The criterion to be minimized.
            "max_depth" : `int` or `None`
                The maximal tree depth.
            "subsample" : `float` or `None`
                The subsampling proportion.
            "max_features" : `int` or `None`
                The maximal number of features to be used in a single split.
            "coef_summary_df" : `pandas.DataFrame`
                The model summary df with the following columns:

                    "Pred_col" : `str`
                        The predictor names.
                    "Feature importance" : `float`
                        The impurity-based feature importance.
                    "Importance rank" : `int`
                        The rank of feature importance.

            "model_type" : `str`
                Equals "tree" for tree models.
    """
    info_dict = create_info_dict_tree(x, y, ml_model, fit_algorithm, pred_cols)
    info_dict = add_model_params_tree(info_dict)
    info_dict = add_model_coef_df_tree(info_dict)
    info_dict["model_type"] = "tree"
    return info_dict


def format_summary_df(summary_df, max_colwidth=20):
    """Formats the summary df for printing.

    Rounds numbers so they look good.
    Formats p-values to make them display the same way as the lm function in R.
    Replaces column names with index if they are too long.

    Parameters
    ----------
    summary_df : `pandas.DataFrame`
        The summary df with the following columns:

            "Pred_col" : `str`
                Names for predictors.
            "Col_index" : `int`
                Indices for predictors. Replaces ``Pred_col`` if predictor names are too long.
            "Estimate" : `float`
                The estimation of coefficients.
            "Std. Err" : `float`, optional
                The standard error for estimations.
            "t value" : `float`, optional
                The t test statistics.
            "z value" : `float`, optional
                The z test statistics. (for glm)
            "Pr(>|t|)" : `float`, optional
                The t test p-values.
            "Pr(>|Z|)" : `float`, optional
                The z test p-value. (for glm)
            "Pr(>)_boot" : `float`, optional
                The bootstrap p-value. (for ridge)
            "sig. code" : `str`, optional
                The significance code for p-values.
            "95%CI" : `list` [ `float` ], optional
                The 95% confidence intervals.

    max_colwidth : `int`
        The maximum length for predictors to be shown in their original name.
        If the maximum length of predictors exceeds this parameter, all
        predictors name will be suppressed and only indices are shown.

    Returns
    -------
    formatted_summary_df : `pandas.DataFrame`
        Formatted summary dataframe.
    """
    summary_df_print = summary_df.copy()
    # Displays column index if the maximum length of column names is too long.
    if summary_df.shape[0] > 0:
        max_colname_length = max([len(name) for name in summary_df["Pred_col"]])
        summary_df_print["Pred_col"] = simplify_pred_cols(summary_df_print["Pred_col"])
        if max_colname_length > max_colwidth:
            summary_df_print["Pred_col"] = [col if len(col) <= max_colwidth
                                            else col[:int((max_colwidth - 3) // 2)]
                                            + "..."
                                            + col[-int((max_colwidth - 3) // 2):]
                                            for col in summary_df_print["Pred_col"]]
        # Rounds columns
        if "Estimate" in summary_df_print.columns:
            summary_df_print["Estimate"] = round_numbers(summary_df_print["Estimate"], 4)
        if "Std. Err" in summary_df_print.columns:
            summary_df_print["Std. Err"] = round_numbers(summary_df_print["Std. Err"], 4)
        if "t value" in summary_df_print.columns:
            summary_df_print["t value"] = round_numbers(summary_df_print["t value"], 4)
        if "z value" in summary_df_print.columns:
            summary_df_print["z value"] = round_numbers(summary_df_print["z value"], 4)
        # Formats p-values
        if "Pr(>|t|)" in summary_df_print.columns:
            summary_df_print["Pr(>|t|)"] = ["<2e-16" if val < 2e-16 else
                                            "{:.2e}".format(val) if val < 1e-3 else
                                            "{0:.3f}".format(val)
                                            for val in summary_df_print["Pr(>|t|)"]]
        if "Pr(>|Z|)" in summary_df_print.columns:
            summary_df_print["Pr(>|Z|)"] = ["<2e-16" if val < 2e-16 else
                                            "{:.2e}".format(val) if val < 1e-3 else
                                            "{0:.3f}".format(val)
                                            for val in summary_df_print["Pr(>|Z|)"]]
        if "Pr(>)_boot" in summary_df_print.columns:
            summary_df_print["Pr(>)_boot"] = ["<2e-16" if val < 2e-16 else
                                              "{:.2e}".format(val) if val < 1e-3 else
                                              "{0:.3f}".format(val)
                                              for val in summary_df_print["Pr(>)_boot"]]
        if "Pr(>)_split" in summary_df_print.columns:
            summary_df_print["Pr(>)_split"] = ["<2e-16" if val < 2e-16 else
                                               "{:.2e}".format(val) if val < 1e-3 else
                                               "{0:.3f}".format(val)
                                               for val in summary_df_print["Pr(>)_split"]]
        # Formats confidence intervals
        if "95%CI" in summary_df_print.columns:
            lbs = round_numbers([ci[0] for ci in summary_df_print["95%CI"]], 4)
            ubs = round_numbers([ci[1] for ci in summary_df_print["95%CI"]], 4)
            summary_df_print["95%CI"] = list(zip(lbs, ubs))
        # Formats feature importance for tree
        if "Feature importance" in summary_df_print.columns:
            summary_df_print["Feature importance"] = round_numbers(summary_df_print["Feature importance"], 4)
    return summary_df_print


def create_title_section():
    """Creates the title section for model summary.

    Returns
    -------
    content : `str`
        Title section.
    """
    content = " Model Summary ".center(80, "=") + "\n\n"
    return content


def create_model_parameter_section(info_dict):
    """Creates the model parameter section for model summary.

    Parameters
    ----------
    info_dict : `dict`
        The dictionary returned by
        `~greykite.algo.common.model_summary_utils.get_info_dict_lm`.
        or
        `~greykite.algo.common.model_summary_utils.get_info_dict_tree`.

    Returns
    -------
    content : `str`
        The model parameter section.
    """
    content = f"Number of observations: {info_dict['n_sample']}"
    content += ",   "
    content += f"Number of features: {info_dict['n_feature']}"
    content += "\n"
    content += f"Method: {info_dict['model']}"
    content += "\n"
    if info_dict["model_type"] == "lm":
        content += f"Number of nonzero features: {info_dict['n_feature_nonzero']}"
        content += "\n"
        if info_dict["fit_algorithm"] == "statsmodels_glm":
            content += f"Family: {info_dict['family']}"
            content += ",   "
            content += f"Link function: {info_dict['link_function']}"
            content += "\n"
        elif info_dict["fit_algorithm"] in ["ridge", "lasso", "lars", "lasso_lars",
                                            "elastic_net", "sgd"]:
            content += f"Regularization parameter: {round_numbers(info_dict['alpha'], 4)}"
            if info_dict["fit_algorithm"] in ["elastic_net", "sgd"]:
                content += ",   "
                content += f"l1_ratio: {round_numbers(info_dict['l1_ratio'], 4)}"
            content += "\n"
    elif info_dict["model_type"] == "tree":
        content += f"Number of Trees: {info_dict['num_tree']}"
        content += ",   "
        content += f"Criterion: {info_dict['criterion'].upper()}"
        content += "\n"
        content += f"Subsample: {info_dict['subsample']}"
        content += ",   "
        content += f"Max features: {info_dict['max_features']}"
        content += ",   "
        content += f"Max depth: {info_dict['max_depth']}"
        content += "\n"
    content += "\n"
    return content


def create_residual_section(info_dict):
    """Creates the residual section for model summary.

    Parameters
    ----------
    info_dict : `dict`
        The dictionary returned by
        `~greykite.algo.common.model_summary_utils.get_info_dict_lm`.
        or
        `~greykite.algo.common.model_summary_utils.get_info_dict_tree`.

    Returns
    -------
    content : `str`
        The residual section.
    """
    residual_summary = round_numbers(info_dict["residual_summary"], 4)
    content = "Residuals:\n"
    content += "{:>12} {:>12} {:>12} {:>12} {:>12}".format("Min", "1Q", "Median", "3Q", "Max") + "\n"
    content += "{:>12} {:>12} {:>12} {:>12} {:>12}".format(*list(residual_summary)) + "\n"
    content += "\n"
    return content


def create_coef_df_section(info_dict, max_colwidth=20):
    """Creates the coefficient summary df section for model summary.

    Parameters
    ----------
    info_dict : `dict`
        The dictionary returned by
        `~greykite.algo.common.model_summary_utils.get_info_dict_lm`.
        or
        `~greykite.algo.common.model_summary_utils.get_info_dict_tree`.
    max_colwidth : `int`
        The maximum length for predictors to be shown in their original name.
        If the maximum length of predictors exceeds this parameter, all
        predictors name will be suppressed and only indices are shown.

    Returns
    -------
    content : `str`
        The coefficient summary df section.
    """
    content = format_summary_df(info_dict["coef_summary_df"], max_colwidth).to_string(index=False)
    content += "\n"
    if "significance_code_legend" in info_dict.keys():
        content += f"Signif. Code: {info_dict['significance_code_legend']}"
        content += "\n"
    content += "\n"
    return content


def create_significance_section(info_dict):
    """Creates the model sifnificance section for model summary.

    Parameters
    ----------
    info_dict : `dict`
        The dictionary returned by
        `~greykite.algo.common.model_summary_utils.get_info_dict_lm`.
        or
        `~greykite.algo.common.model_summary_utils.get_info_dict_tree`.

    Returns
    -------
    content : `str`
        The model significance section.
    """
    if info_dict["model_type"] == "lm":
        content = f"Multiple R-squared: {round_numbers(info_dict['r2'], 4)},"
        content += "   "
        content += f"Adjusted R-squared: {round_numbers(info_dict['r2_adj'], 4)}"
        content += "\n"
        content += f"F-statistic: {round_numbers(info_dict['f_value'], 5)} " \
                   f"on {int(info_dict['df_ssr'])} and" \
                   f" {int(info_dict['df_sse'])} DF,"
        content += "   "
        content += f"p-value: {round_numbers(info_dict['f_p_value'], 4)}"
        content += "\n"
        content += f"Model AIC: {round_numbers(info_dict['aic'], 5)},"
        content += "   "
        content += f"model BIC: {round_numbers(info_dict['bic'], 5)}"
        content += "\n\n"
        return content
    else:
        return None


def create_warning_section(info_dict):
    """Creates the warning section for model summary.

    The following warnings are possible to be included:

        Condition number is too big.
        F-ratio and p-value on regularized methods.
        R-squared and F-ratio in glm.
        Zero coefficients in non-sparse regression methods due to degenerate columns.

    Parameters
    ----------
    info_dict : `dict`
        The dictionary returned by
        `~greykite.algo.common.model_summary_utils.get_info_dict_lm`.
        or
        `~greykite.algo.common.model_summary_utils.get_info_dict_tree`.

    Returns
    -------
    content : `str`
        The model warnings section.
    """
    if info_dict["model_type"] == "lm":
        content = ""
        if info_dict["condition_number"] > 1000:
            content += f"WARNING: the condition number is large, {'{:.2e}'.format(info_dict['condition_number'])}. " \
                       f"This might indicate that there are strong multicollinearity " \
                       f"or other numerical problems."
            content += "\n"
        if info_dict["fit_algorithm"] in ["ridge", "lasso", "lars", "lasso_lars",
                                          "sgd", "elastic_net"]:
            content += "WARNING: the F-ratio and its p-value on regularized methods might be misleading, " \
                       "they are provided only for reference purposes."
            content += "\n"
        if info_dict["fit_algorithm"] in ["statsmodels_glm"]:
            content += "WARNING: the R-squared and F-statistics on glm might be misleading, " \
                       "they are provided only for reference purposes."
            content += "\n"
        if (len(info_dict["nonzero_index"]) < info_dict["n_feature"]
            and
            info_dict["fit_algorithm"] in ["linear", "ridge", "statsmodels_ols",
                                           "statsmodels_wls", "statsmodels_gls",
                                           "statsmodels_glm"]):
            content += "WARNING: the following columns have estimated coefficients equal to zero, " \
                       f"while {info_dict['fit_algorithm']} is not supposed to have zero estimates. " \
                       f"This is probably because these columns are degenerate in the design matrix. " \
                       f"Make sure these columns do not have constant values." + "\n" +\
                       f"{[col for i, col in enumerate(info_dict['pred_cols']) if i not in info_dict['nonzero_index']]}"
            content += "\n"
        if len(info_dict["degenerate_index"]) > 1:
            content += "WARNING: the following columns are degenerate, do you really want to include them in your model? " \
                       "This may cause some of them to show unrealistic significance. Consider using the `drop_degenerate` transformer." + "\n" +\
                       f"{[col for i, col in enumerate(info_dict['pred_cols']) if i in info_dict['degenerate_index']]}"
            content += "\n"
        if content == "":
            content = None
    else:
        content = None
    return content


def print_summary(info_dict, max_colwidth=20):
    """Creates the content for printing the summary.

    Parameters
    ----------
    info_dict : `dict`
        The output summary dictionary from
        `~greykite.algo.common.model_summary.ModelSummary._get_summary`,
        which is originally from
        `~greykite.algo.common.model_summary_utils.get_info_dict_lm`
        or
        `~greykite.algo.common.model_summary_utils.get_info_dict_tree`.
    max_colwidth : `int`
        The maximum length for predictors to be shown in their original name.
        If the maximum length of predictors exceeds this parameter, all
        predictors name will be suppressed and only indices are shown.

    Returns
    -------
    summary_content : `str`
        The content to be printed.
    """
    content = create_title_section()
    model_param_section = create_model_parameter_section(info_dict)
    if model_param_section is not None:
        content += model_param_section
    residual_section = create_residual_section(info_dict)
    if residual_section is not None:
        content += residual_section
    coef_df_section = create_coef_df_section(info_dict, max_colwidth)
    if coef_df_section is not None:
        content += coef_df_section
    signif_section = create_significance_section(info_dict)
    if signif_section is not None:
        content += signif_section
    warning_section = create_warning_section(info_dict)
    if warning_section is not None:
        content += warning_section
    return content



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Calculates uncertainty intervals from the conditional
empirical distribution of the residual.
"""

import warnings

import numpy as np
import pandas as pd

from greykite.algo.uncertainty.conditional.dataframe_utils import limit_tuple_col
from greykite.algo.uncertainty.conditional.dataframe_utils import offset_tuple_col
from greykite.algo.uncertainty.conditional.estimate_distribution import estimate_empirical_distribution
from greykite.algo.uncertainty.conditional.normal_quantiles import normal_quantiles_df
from greykite.common.constants import ERR_STD_COL


def conf_interval(
        df,
        value_col,
        residual_col=None,
        conditional_cols=None,
        quantiles=[0.005, 0.025, 0.975, 0.995],
        quantile_estimation_method="normal_fit",
        sample_size_thresh=5,
        small_sample_size_method="std_quantiles",
        small_sample_size_quantile=0.95,
        min_admissible_value=None,
        max_admissible_value=None):
    """A function to calculate confidence intervals (ci) for values given
    in ``value_col``
    for each slice of data (given in ``conditional_cols``)
    using approximate distributions estimated via
    ``estimate_empirical_distribution``.
    The variability of the CI's either come from "value_col" itself or
    alternatively come from the ``residual_col`` if provided.

    We allow for calculating as many quantiles as needed (specified by `quantiles`)
    as opposed to only two quantiles representing a typical ci interval.

    Two options are available for method of calculation of quantiles for
    each slice,

        - for each slice a confidence interval is calculated
            simply using empirical quantiles
        - using a normal distribution fit.

    There are two main possibilities:

        - ``residual_col`` is not provided and values in value_col are used
            directly to calculate quantiles by using the distribution of the
            values in each slices
        - ``residual_col`` is provided, we calculate quantiles
            for the residuals distribution for each slice and then
            offset the quantiles by the value given in value_col.
            In that case we use a fixed_mean of zero when constructing quantiles
            for the residuals. This is done so that our predicted values given
            in value_col are not perturbed as they might be coming from a much
            more complex fitting model taking into account many more variables
            as compared with conditional_cols

    Parameters
    ----------
    df : `pandas.Dataframe`
        The dataframe with the needed columns:

            - value_col,
            - conditional_cols,
            - residual_col (optional column)

    value_col : str
        The column containing the values for the variable for which confidence
        interval is needed
    residual_col : str
        If a residual column is given, quantiles will be built for the residual
        values and the interval is then offset using the value given in
        value_col itself
    conditional_cols : `list` [`str`]
        These columns are used to slice the data first then calculate quantiles
        for each slice
    quantiles : `list` [`float`]
        The quantiles calculated for each slice.
        These quantiles can be then used to construct the desired CIs.
        The default values [0.005, 0.025, 0.0975, 0.995] can be used to construct
        99 and 95 percent CIs.
    quantile_estimation_method : `str`
        There are two options implemented for the quantile estimation method
        (conditional on slice):

            - "normal_fit": it uses the standard deviation of the values in each
              slice to compute normal distribution quantiles
            - "ecdf": it uses the values directly to calculate sample quantiles

    sample_size_thresh : int
        The minimum sample size for each slice where we allow for using the conditional
        distribution (conditioned on the "conditional_cols" argument).
        If sample size for that slice is smaller than this,
        we fall back to a fallback method
    small_sample_size_method : str
        The method to use for slices with small sample size

            - "std_quantile" method is implemented and it looks at the response
              std for each slice with
              sample size >= "sample_size_thresh"
              and takes the row which has  its std being closest
              to "small_sample_size_quantile" quantile.
              It assigns that row to act as fall-back for calculating conf
              intervals.

    min_admissible_value : Union[float, double, int]
        This is the lowest admissible value for the obtained ci limits
        and any value below this will be mapped back to this value.
    max_admissible_value : Union[float, double, int]
        This is the highest admissible value for the obtained ci limits
        and any higher value will be mapped back to this value.

    Returns
    -------
    uncertainty_model : `dict`
        Dict with following items (main component is the ``predict`` function).

            - "ecdf_df": pandas.DataFrame
                ecdf_df generated by "estimate_empirical_distribution"
            - "ecdf_df_overall": pandas.DataFrame
                ecdf_df_overall generated by "estimate_empirical_distribution"
            - "ecdf_df_fallback": pandas.DataFrame
                ecdf_df_fallback, a fall back data to get the CI quantiles
                when the sample size for that slice is small or that slice
                is unobserved in that case.

                    - if small_sample_size_method = "std_quantiles",
                      we use std quantiles to pick a slice which has a std close
                      to that quantile and fall-back to that slice.
                    - otherwise we fallback to "ecdf_overall"

            - "predict": callable
                it can apply to new_df and add quantiles
                the new column added has the extra column with name
                "{value_col}_quantile_summary"
                as well as the input slices given in "conditional_cols"

    """
    is_residual_based = (residual_col is not None)
    distribution_col = residual_col if is_residual_based else value_col
    std_col = f"{distribution_col}_std"
    sample_size_col = f"{distribution_col}_count"

    model_dict = estimate_empirical_distribution(
        df=df,
        value_col=distribution_col,
        quantile_grid_size=None,
        quantiles=quantiles,
        conditional_cols=conditional_cols)

    ecdf_df = model_dict["ecdf_df"]
    ecdf_df_overall = model_dict["ecdf_df_overall"]
    ecdf_df_fallback = ecdf_df_overall.copy()

    # two methods are implemented: ecdf; normal_fit.
    # we re-assign the quantile_summary_col if
    # the quantile_estimation_method is normal_fit
    if quantile_estimation_method == "ecdf":
        quantile_summary_col = f"{distribution_col}_quantile_summary"
    elif quantile_estimation_method == "normal_fit":
        quantile_summary_col = "normal_quantiles"
        if is_residual_based:
            mean_col = None
            fixed_mean = 0.0
        else:
            mean_col = value_col
            fixed_mean = None

        ecdf_df = normal_quantiles_df(
            df=ecdf_df,
            std_col=std_col,
            mean_col=mean_col,
            fixed_mean=fixed_mean,
            quantiles=quantiles)

        ecdf_df_fallback = normal_quantiles_df(
            df=ecdf_df_fallback,
            std_col=std_col,
            mean_col=mean_col,
            fixed_mean=fixed_mean,
            quantiles=quantiles)
    else:
        raise NotImplementedError(
                f"CI calculation method {quantile_estimation_method} is not either of: normal_fit; ecdf")

    # handling slices with small sample size
    # if a method is provided via the argument "small_sample_size_method" then it is used here
    # the idea is to take a relatively high volatility
    # when the new point does not have enough (as specified by "sample_size_thresh")
    # similar points in the past
    fall_back_for_all = False
    if small_sample_size_method == "std_quantiles":
        ecdf_df_large_ss = ecdf_df.loc[ecdf_df[sample_size_col] >= sample_size_thresh].reset_index(drop=True)
        assert set(ecdf_df_large_ss.columns).intersection(["std_quantile", "std_quantile_diff"]) == set(), (
            "column names: std_quantile, std_quantile_diff should not appear in ecdf_df")
        if len(ecdf_df_large_ss) == 0:
            warnings.warn("No slice had sufficient sample size. We fall back to the overall distribution.")
            # If ``ecdf_df_large_ss`` is empty it means we do not have any sufficient
            # samples for any slices.
            # Therefore we have to fall back in all cases and we set ``ecdf_df``
            # to ``ecdf_df_fall_back``
            ecdf_df = ecdf_df_fallback
            fall_back_for_all = True
        else:
            ecdf_df_large_ss["std_quantile"] = np.argsort(ecdf_df_large_ss[std_col]) / ecdf_df_large_ss.shape[0]
            # Calculates the distance between "std_quantile" column values and ``small_sample_size_quantile``
            ecdf_df_large_ss["std_quantile_diff"] = abs(ecdf_df_large_ss["std_quantile"] - small_sample_size_quantile)
            # Chooses the row with closes value in "std_quantile" column to ``small_sample_size_quantile``
            # Note the resulting dataframe below ``ecdf_df_fallback`` will have one row
            ecdf_df_fallback = ecdf_df_large_ss.loc[[ecdf_df_large_ss["std_quantile_diff"].idxmin()]]
            del ecdf_df_fallback["std_quantile"]
            del ecdf_df_fallback["std_quantile_diff"]
            del ecdf_df_large_ss["std_quantile"]
            del ecdf_df_large_ss["std_quantile_diff"]
            # we re-assign ecdf_df by removing the combinations with small sample size
            # this is done so that in predict phase those values are not populated from
            # small sample sizes and use ``ecdf_fallback``
            ecdf_df = ecdf_df_large_ss
    elif small_sample_size_method is not None:
        raise NotImplementedError(
            f"small_sample_size_method {small_sample_size_method} is not implemented.")

    return {
        "ecdf_df": ecdf_df,
        "ecdf_df_overall": ecdf_df_overall,
        "ecdf_df_fallback": ecdf_df_fallback,
        "value_col": value_col,
        "min_admissible_value": min_admissible_value,
        "max_admissible_value": max_admissible_value,
        "conditional_cols": conditional_cols,
        "std_col": std_col,
        "quantile_summary_col": quantile_summary_col,
        "fall_back_for_all": fall_back_for_all,
        "is_residual_based": is_residual_based}


def predict_ci(
        new_df,
        ci_model):
    """It applies on a dataframe (``new_df``) and attaches the quantiles needed.

    :param new_df: pd.Dataframe
        A dataframe with ``value_col`` column as mandatory and
        ``conditional_cols`` as optional depending on
        how the function ``conf_interval`` is called.
    :parame ci_model: dict
        Returned CI model from ``conf_interval``.
    :return: pd.Dataframe
        A dataframe which includes ``new_df`` and new columns containing
        the quantiles.
    """

    ecdf_df = ci_model["ecdf_df"]
    ecdf_df_fallback = ci_model["ecdf_df_fallback"]
    value_col = ci_model["value_col"]
    min_admissible_value = ci_model["min_admissible_value"]
    max_admissible_value = ci_model["max_admissible_value"]
    conditional_cols = ci_model["conditional_cols"]
    std_col = ci_model["std_col"]
    quantile_summary_col = ci_model["quantile_summary_col"]
    fall_back_for_all = ci_model["fall_back_for_all"]
    is_residual_based = ci_model["is_residual_based"]

    # copy ``new_df`` so that input df to predict is not altered
    new_df = new_df.copy().reset_index(drop=True)
    new_df["temporary_overall_dummy"] = 0
    ecdf_df_fallback_dummy = ecdf_df_fallback.copy()
    ecdf_df_fallback_dummy["temporary_overall_dummy"] = 0
    new_df_fallback = pd.merge(
        new_df,
        ecdf_df_fallback_dummy,
        on=["temporary_overall_dummy"],
        how="left")

    if conditional_cols is None or fall_back_for_all:
        new_df_conditional = new_df_fallback.copy()
    else:
        new_df_conditional = pd.merge(
            new_df,
            ecdf_df,
            on=conditional_cols,
            how="left")
    del new_df_conditional["temporary_overall_dummy"]

    # when we have missing in the grouped case (which can happen if a level
    # in ``match_cols`` didn't appear in train dataset)
    # we fall back to the overall case
    for col in [quantile_summary_col, std_col]:
        na_index = new_df_conditional[col].isnull()
        new_df_conditional.loc[na_index, col] = (
            new_df_fallback.loc[na_index, col])

    # offsetting the values in ``value_col`` by residual quantiles
    # if ``is_residual_based``
    value_quantile_summary_col = f"{value_col}_quantile_summary"
    if is_residual_based:
        new_df_conditional[value_quantile_summary_col] = offset_tuple_col(
            df=new_df_conditional,
            offset_col=value_col,
            tuple_col=quantile_summary_col)

    new_df_conditional = limit_tuple_col(
        df=new_df_conditional,
        tuple_col=value_quantile_summary_col,
        lower=min_admissible_value,
        upper=max_admissible_value)

    # only returning needed cols
    returned_cols = [value_quantile_summary_col, std_col]
    if conditional_cols is not None:
        returned_cols = conditional_cols + returned_cols

    pred_df = new_df_conditional[returned_cols]
    pred_df.rename(columns={std_col: ERR_STD_COL}, inplace=True)

    return pred_df



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Creates a forecast using similar time points in the past."""
import warnings

import numpy as np
import pandas as pd

from greykite.common.features.timeseries_features import build_time_features_df
from greykite.common.features.timeseries_features import convert_date_to_continuous_time
from greykite.common.time_properties import describe_timeseries


def forecast_similarity_based(
        df,
        time_col,
        value_cols,
        agg_method,
        agg_func=None,
        grid_size_str=None,
        match_cols=[],
        origin_for_time_vars=None,
        recent_k=1):
    """Fits a basic forecast model which is aggregate based. As an example for an hourly time series we
    can assign the value of the most recent three weeks at the same time of the week as forecast.
    This works for multiple responses passed as a list in "value_cols".
    Also we do not require the timestamps to be regular and for example dataframe can have missing
    timestamps.

    :param df: the dataframe which includes the training data, the value_cols and match_cols if needed
    :param time_col: the column of the dataframe which includes the timestamps of the series
    :param value_cols: the response columns for which forecast is desired
    :param agg_method: a string which specifies the aggregation method. Options are
        "mean": the mean of the values at timestamps which match with desired time on match_cols
        "median": the median of the values at timestamps which match with desired time on match_cols
        "min": the min of the values at timestamps which match with desired time on match_cols
        "max": the max of the values at timestamps which match with desired time on match_cols
        "most_recent": the mean of "recent_k" (given in last argument of the function)
        values matching with desired time on match_cols
    :param agg_func: the aggregation function needed for aggregating w.r.t "match_cols".
        This is needed if agg_method is not available.
    :param grid_size_str: the expected time increment. If not provided it is inferred.
    :param match_cols: the variables used for grouping in aggregation
    :param origin_for_time_vars: the time origin for continuous time variables
    :param recent_k: the number of most recent timestamps to consider for aggregation

    :return: A dictionary consisting of a "model" and a "predict" function.
        The "model" object is simply a dictionary of two items (dataframes).
        First item is an aggregated dataframe of "value_cols" w.r.t "match_cols": "pred_df"
        Second item is an aggregated dataframe across all data: "pred_df_overall".
        This is useful if a level in match_cols didn't appear in train dataset,
        therefore in that case we fall back to overall prediction
        The "predict" is a function which performs prediction for new data
        The "predict_n" is a function which predicts the future for any given number of steps specified
    """
    # This is only needed for predict_n function.
    # But it makes sense to do this only once for computational efficiency
    timeseries_info = describe_timeseries(df, time_col)
    if grid_size_str is None:
        grid_size = timeseries_info["median_delta"]
        grid_size_str = str(int(grid_size / np.timedelta64(1, "s"))) + "s"
    max_timestamp = timeseries_info["max_timestamp"]

    # a dictionary of methods and their corresponding aggregation function
    agg_method_func_dict = {
        "mean": (lambda x: np.mean(x)),
        "median": (lambda x: np.median(x)),
        "min": (lambda x: np.min(x)),
        "max": (lambda x: np.max(x)),
        "most_recent": (lambda x: np.mean(x[-recent_k:]))
    }

    if agg_func is None:
        if agg_method not in agg_method_func_dict.keys():
            raise Exception(
                "The aggregation method you specified is not implemented." +
                "These are the available methods: mean, median, min, max, most_recent")
        agg_func = agg_method_func_dict[agg_method]

    # sets default origin so that "ct1" feature from "build_time_features_df" starts at 0 on training start date
    if origin_for_time_vars is None:
        origin_for_time_vars = convert_date_to_continuous_time(df[time_col][0])

    # we calculate time features here which might appear in match_cols and not available in df by default
    def add_time_features(df):
        time_df = build_time_features_df(dt=df[time_col], conti_year_origin=origin_for_time_vars)
        for col in match_cols:
            if col not in df.columns:
                df[col] = time_df[col].values
        return df

    df = add_time_features(df=df)

    # for value_col in value_cols:
    # aggregate w.r.t. given columns in match_cols
    agg_dict = {value_col: agg_func for value_col in value_cols}
    # we create a coarse prediction by simply aggregating every value column globally
    # this is useful for cases where no matching data is available for a given timestamp to be forecasted
    pred_df_overall = df.groupby([True]*len(df), as_index=False).agg(agg_dict)
    if len(match_cols) == 0:
        pred_df = pred_df_overall
    else:
        pred_df = df.groupby(match_cols, as_index=False).agg(agg_dict)

    model = {
        "pred_df": pred_df,
        "pred_df_overall": pred_df_overall}

    def predict(new_df, new_external_regressor_df=None):
        """Predicts for new dataframe (new_df) using the fitted model.
        :param new_df: a dataframe of new data which must include the time_col and match_cols
        :param new_external_regressor_df: a regressor dataframe if needed
        :return: new_df is augmented with predictions for value_cols and returned
        """
        new_df = new_df.copy(deep=True)
        new_df = add_time_features(df=new_df)
        if new_external_regressor_df is not None:
            new_df = pd.concat([new_df, new_external_regressor_df])

        # if the response columns appear in the new_df columns, we take them out to prevent issues in aggregation
        for col in value_cols:
            if col in new_df.columns:
                warnings.warn(f"{col} is a response column and appeared in new_df. Hence it was removed.")
                del new_df[col]

        new_df["temporary_overall_dummy"] = 0
        pred_df_overall["temporary_overall_dummy"] = 0

        new_df_grouped = pd.merge(new_df, pred_df, on=match_cols, how="left")
        new_df_overall = pd.merge(new_df, pred_df_overall, on=["temporary_overall_dummy"], how="left")

        # when we have missing in the grouped case (which can happen if a level in match_cols didn't appear in train dataset)
        # we fall back to the overall case
        for col in value_cols:
            new_df_grouped.loc[new_df_grouped[col].isnull(), col] = new_df_overall.loc[new_df_grouped[col].isnull(), col]

        del new_df_grouped["temporary_overall_dummy"]

        return new_df_grouped

    def predict_n(fut_time_num, new_external_regressor_df=None):
        """This is the forecast function which can be used to forecast.
        It accepts extra predictors if needed in the form of a dataframe: new_external_regressor_df.
        :param fut_time_num: number of needed future values
        :param new_external_regressor_df: extra predictors if available
        """
        # we create the future time grid
        date_list = pd.date_range(
            start=max_timestamp + pd.Timedelta(grid_size_str),
            periods=fut_time_num,
            freq=grid_size_str).tolist()

        fut_df = pd.DataFrame({time_col: date_list})
        return predict(fut_df, new_external_regressor_df=new_external_regressor_df)

    return {"model": model, "predict": predict, "predict_n": predict_n}
